package anune;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import anune.util.Utils;
import java.util.Arrays;


public class ANN implements Serializable {
	
	private final ArrayList<Layer> layers;
	private final int configuration[];
	private final int hiddenConfiguration[];
	transient private float input[]; // No need to serialize this
	
	private int nrOfWeights;
	private int nrOfBiases;
	private int nrOfTaus;
	private int nrOfGains;

	
	public ANN(int[] configuration) {
		this.configuration = new int[configuration.length];
		System.arraycopy(configuration, 0, this.configuration, 0, configuration.length);
		
		this.hiddenConfiguration = new int[configuration.length - 2];
		System.arraycopy(configuration, 1, this.hiddenConfiguration, 0, this.hiddenConfiguration.length);
		
		this.layers = new ArrayList<>(configuration.length);
		
		layers.add(new InputLayer(configuration[0], this));
		for (int i = 1; i < configuration.length; i++) {
			layers.add(new Layer(configuration[i], layers.get(i-1)));
		}
		
		// Set counts
		updateCounts();
	}
	
	public ANN(ANN orig) {
		this(orig.configuration);
		for (int i = 0; i < layers.size(); i++) {
			for (int j = 0; j < layers.get(i).neurons.size(); j++) {
				System.arraycopy(orig.getLayers().get(i).neurons.get(j).sameLayerWeights, 0, layers.get(i).neurons.get(j).sameLayerWeights, 0, layers.get(i).neurons.get(j).sameLayerWeights.length);
				System.arraycopy(orig.getLayers().get(i).neurons.get(j).weights, 0, layers.get(i).neurons.get(j).weights, 0, layers.get(i).neurons.get(j).weights.length);
				this.layers.get(i).neurons.get(j).tau = orig.layers.get(i).neurons.get(j).tau;
				this.layers.get(i).neurons.get(j).gain = orig.layers.get(i).neurons.get(j).gain;
				this.layers.get(i).neurons.get(j).bias = orig.layers.get(i).neurons.get(j).bias;
			}
		}
	}
	
	private void updateCounts(){
		nrOfWeights = 0;
		nrOfTaus = 0;
		nrOfGains = 0;
		nrOfBiases = 0;
		
		for (Layer l : layers){
			for (Neuron n : l.neurons){
				nrOfWeights += n.weights.length + n.sameLayerWeights.length;
			}
			nrOfTaus += l.neurons.size();
			nrOfGains += l.neurons.size();
			nrOfBiases += l.neurons.size();
		}
	}
	
	
	public float[] getInput() {
		return input;
	}
	
	
	public float[] run(float[] input) {
//		System.out.println(Arrays.toString(input));
		if (input.length != configuration[0])
			throw new IllegalArgumentException(String.format("Wrong number of inputs to ANN. Expected %d, got %d.",
					configuration[0],
					input.length)
			);
		
		//System.arraycopy(input, 0, this.input, 0, input.length);
		this.input = input;
		
		for (Layer l : layers)
			l.calcOutput();
		
		return layers.get(layers.size()-1).output;
	}
	
	/**
	 * Updates the weights of the ANN
	 * Takes in a double array of weights and updates neuron by neuron
	 * The list should be structured:
	 *			
	 *			weights[0]				  : n.weights[0]
	 *			weights[1]				  : n.weights[1]
	 *					:
	 *					:
	 *					:
	 *			weights[weights.length]   : n.sameLayerWeights[0]
	 *			weights[weights.length+1] : n.sameLayerWeights[1]
	 *					:
	 *					:
	 *					:
	 * 
	 * @param weights 
	 */
	public void updateWeights(float[] weights) {
		if (weights.length != nrOfWeights)
			throw new IllegalArgumentException(String.format("Wrong number of weights in input. Expected %d, got %d.",
					nrOfWeights,
					weights.length)
			);
		int counter = 0;
		for (Layer l : layers){
			for (Neuron n : l.neurons){
				// Updates outgoing weights
				for (int i = 0; i < n.weights.length; i++){
					n.weights[i] = weights[counter++];
				}
				// Updates same layer weigths
				for (int i = 0; i < n.sameLayerWeights.length; i++){
					n.sameLayerWeights[i] = weights[counter++];
				}
			}
		}
	}
	
	public int getWeightsCount(){
		return this.nrOfWeights;
	}
	/**
	 * Updates the tau values in each neuron
	 * 
	 * @param taus 
	 */
	public void updateTaus(float[] taus) {
		if (taus.length != nrOfTaus)
			throw new IllegalArgumentException(String.format("Wrong number of tau values in input. Expected %d, got %d.",
					nrOfTaus,
					taus.length)
			);
		// Updates the tau values
		int counter = 0;
		for (Layer l : layers){
			for (Neuron n : l.neurons){
				n.tau = taus[counter++];
			}
		}
	}
	public int getTausCount(){
		return this.nrOfTaus;
	}
	
	/**
	 * Updates the gain values in each neuron
	 * 
	 * @param gains 
	 */
	public void updateGains(float[] gains) {
		if (gains.length != nrOfGains)
			throw new IllegalArgumentException(String.format("Wrong number of gain values in input. Expected %d, got %d.",
					nrOfGains,
					gains.length)
			);
		// Updates the gain values
		int counter = 0;
		for (Layer l : layers){
			for (Neuron n : l.neurons){
				n.gain = gains[counter++];
			}
		}
	}
	public int getGainsCount(){
		return this.nrOfGains;
	}
	
	/**
	 * Updates the bias values in each neuron
	 * 
	 * @param biases 
	 */
	public void updateBias(float[] biases) {
		if (biases.length != nrOfBiases)
			throw new IllegalArgumentException(String.format("Wrong number of bias values in input. Expected %d, got %d.",
					nrOfBiases,
					biases.length)
			);
		// Updates the bias values
		int counter = 0;
		for (Layer l : layers){
			for (Neuron n : l.neurons){
				n.bias = biases[counter++];
			}
		}
	}
	public int getBiasesCount(){
		return this.nrOfBiases;
	}
	
	/**
	 * Adds a neuron to the input
	 * 
	 * Increases the configuration to accept 1 more input
	 */
	public void addInputNeuron() {
		configuration[0]++;
		Layer nextLayer = layers.get(1);
		// Modify next layer weigths
		for (Neuron n : nextLayer.neurons){
			float[] newWeights = new float[n.weights.length+1];
			System.arraycopy(n.weights, 0, newWeights, 0, newWeights.length);
			newWeights[newWeights.length-1] = Utils.randomFloat()*2 - 1;
		}
		updateCounts();
	}
	
	/**
	 * Adds a neuron to the output
	 * 
	 * Creates a new neuron that it puts in the last layer
	 * Also adds 1 sameLayerWeight to all other output neurons
	 */
	public void addOutputNeuron() {
		configuration[configuration.length-1]++;
		Layer outputLayer = layers.get(layers.size()-1);
		// Modify same layer weigths
		for (Neuron n : outputLayer.neurons){
			float[] newSameLayerWeights = new float[n.sameLayerWeights.length+1];
			System.arraycopy(n.sameLayerWeights, 0, newSameLayerWeights, 0, newSameLayerWeights.length);
			newSameLayerWeights[newSameLayerWeights.length-1] = Utils.randomFloat()*2 - 1;
		}
		// Add the neuron
		Neuron n = new Neuron(layers.get(layers.size()-2).neurons.size(), outputLayer.neurons.size()+1);
		outputLayer.neurons.add(n);
		
		updateCounts();
		
	}
	
	
	public void serialize(String filename) throws IOException {
		ObjectOutputStream oos = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(filename)));
		oos.writeObject(this);
		oos.close();
	}
	
	
	public int[] getConfiguration() {
		return configuration;
	}
	
	
	public int[] getHiddenConfiguration() {
		return hiddenConfiguration;
	}
	
	public ArrayList<Layer> getLayers(){
		return this.layers;
	}
			
			
	public static ANN deserialize(String filename) throws IOException, ClassNotFoundException {
		ObjectInputStream ois = new ObjectInputStream(new BufferedInputStream(new FileInputStream(filename)));
		ANN ann = (ANN) ois.readObject();
		ois.close();
		return ann;
	}
	
	
}
