package creativebehaviour;

import creativebehaviour.ea.*;
import creativebehaviour.ents.*;
import creativebehaviour.graphics.MainFrame;
import creativebehaviour.graphics.evolvedlistgui.EvolvedListGUI;
import creativebehaviour.models.ann.ANNCreature;
import creativebehaviour.models.Creature;
import creativebehaviour.models.DummyCreature;
import creativebehaviour.models.physical.TreeGenotype;
import creativebehaviour.models.ann.WeightsGenotype;
import creativebehaviour.simulation.World;
import creativebehaviour.simulation.WorldFactory;
import creativebehaviour.simulation.WorldType;
import creativebehaviour.util.*;
import creativebehaviour.util.io.CrebeIO;
import creativebehaviour.util.profiling.Timer;
import eval.ea.EA;
import eval.ea.EAResults;
import eval.ea.OnChangeListener;
import eval.ea.OnChangedEvent;
import eval.models.Population;
import eval.util.Config;
import java.awt.HeadlessException;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.Locale;
import java.util.Objects;
import java.util.Random;
import java.util.function.Consumer;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;


public class CreativeBehaviour {
	
	private static final String[] PRESENTATION_CREATURE =
	{
	};
	
	public static Config crebeCfg;
	public static Config creatureCfg;
	public static Config weightsCfg;
	public static Config neatCfg;
	public static Config fuzzyCfg;
	
	public static WorldType worldType;
	public static BrainEA.Method brainMethod;
	public static RunMode runMode = RunMode.NONE;
	public static int runs;
	public static int currentRun;
	public static boolean dumpFitness;
	public static String runName;
	public static String designedCreatureFileName;
	
	public static boolean DEBUG = false;
	public static boolean HEADLESS = false;
	
	private static final Object CURRENT_BEST_LOCK = new Object();
	private static double currentBestFitness = Double.NEGATIVE_INFINITY;
	
	private static EvolvedListGUI evolvedGui;
	public static boolean aborted = false;
	
	
	public enum RunMode {
		DESIGNED, FULL, NONE, PRESENTATION;
	}
	
	
	static {
		/*
		Read configuration files.
		*/
		try {
			try {
				crebeCfg    = new Config("crebe.properties");
			} catch (IOException e) {
				createConfigFile("crebe.properties");
				crebeCfg = new Config("crebe.properties");
			}
			
			try {
				creatureCfg = new Config("creature_ea.properties");
			} catch (IOException e) {
				createConfigFile("creature_ea.properties");
				creatureCfg = new Config("creature_ea.properties");
			}
			
			try {
				weightsCfg  = new Config("ctrnn_ea.properties");
			} catch (IOException e) {
				createConfigFile("ctrnn_ea.properties");
				weightsCfg = new Config("ctrnn_ea.properties");
			}
			
			try {
				neatCfg = new Config("neat_ea.properties");
			} catch (IOException e) {
				createConfigFile("neat_ea.properties");
				neatCfg = new Config("neat_ea.properties");
			}
			
			try {
				fuzzyCfg = new Config("fuzzy_ea.properties");
			} catch (IOException e) {
				createConfigFile("fuzzy_ea.properties");
				fuzzyCfg = new Config("fuzzy_ea.properties");
			}
		} catch (Exception e) {
			String msg = "Error reading configuration files. " + e.toString() + "\nApplication will now exit.";
			System.err.println(msg);
			e.printStackTrace(System.err);
			try {
				JOptionPane.showMessageDialog(null, msg, "Config file error", JOptionPane.ERROR_MESSAGE);
			} catch (Exception ex) {
			}
			System.exit(1);
		}
	}
	
	
	private static void createConfigFile(String name) throws IOException {
		BufferedReader br = new BufferedReader(new InputStreamReader(
				CreativeBehaviour.class.getResourceAsStream("/creativebehaviour/defaultcfg/" + name)));
		BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(name, false)));
		String line = br.readLine();
		while (line != null) {
			bw.write(line);
			bw.write("\n");
			line = br.readLine();
		}
		br.close();
		bw.close();
	}
	
	
	private static void initTheme() {
        try {
            for (javax.swing.UIManager.LookAndFeelInfo info : javax.swing.UIManager.getInstalledLookAndFeels()) {
                if ("Windows".equals(info.getName())) {
                    javax.swing.UIManager.setLookAndFeel(info.getClassName());
                    break;
                }
            }
        } catch (Exception ex) {}
    }
	
	
	public static void main(String[] args) throws Exception {
		System.out.println("Creative Creature Behaviour (CreBe)");
		System.out.println("Written by Bjørnar W. Alvestad & Endre Larsen");
		System.out.println("for the Master's thesis \"Creative Behaviour in Evolving Agents\", supervisor: Björn Gambäck.");
		System.out.println();
		System.out.println("Department of Computer and Information Science");
		System.out.println("Faculty of Information Technology, Mathematics and Electrical Engineering");
		System.out.println("Norwegian University of Science and Technology");
		System.out.println();
		System.out.println();
		System.out.println();
		
		/*
		Parse args.
		*/
		ArgumentParser argParser = new ArgumentParser(args);
		if (argParser.numArguments() > 0) {
			System.out.println("Arguments:");
			argParser.printAllOptions();
		} else {
			System.out.println("Usage:");
			System.out.println("\t-brain <brain_method>, possible values: " + Arrays.toString(BrainEA.Method.values()).replaceAll("\\[|\\]", ""));
			System.out.println("\t-designed <creature_txt_file>");
			System.out.println("\t-full");
			System.out.println("\t-debug");
			System.out.println("\t-headless");
			System.out.println("\t-seed");
			System.out.println("\t-runs <number_of_runs>");
			System.out.println("\t-dumpfitness <run_name>");
			System.out.println("\t-presentation");
		}
		System.out.println();
		
		DEBUG = argParser.isPresent("-debug");
		HEADLESS = argParser.isPresent("-headless");
		
		System.out.println("Debug: " + DEBUG);
		System.out.println("Headless: " + HEADLESS);
		
		/*
		Parse runs and dumpfitness
		*/
		runs = argParser.getInt("-runs", 1);
		dumpFitness = argParser.isPresent("-dumpfitness");
		runName = argParser.getString("-dumpfitness", "run");
		System.out.println("Runs: " + runs);
		System.out.println("Dump fitness: " + dumpFitness);
		System.out.println("Run-name: " + runName);
		
		/*
		Parse brain method.
		*/
		String argBrainMethod = argParser.getString("-brain");
		if (argBrainMethod != null) {
			for (BrainEA.Method m : BrainEA.Method.values()) {
				if (m.name().equalsIgnoreCase(argBrainMethod)) {
					brainMethod = m;
					break;
				}
			}
		}
		System.out.println("Brain method: " + (brainMethod == null ? "none" : brainMethod.name().toLowerCase()));
		
		/*
		Parse run-mode.
		*/
		if (argParser.isPresent("-designed")) runMode = RunMode.DESIGNED;
		if (argParser.isPresent("-full")) runMode = RunMode.FULL;
		if (argParser.isPresent("-presentation")) {
			runMode = RunMode.PRESENTATION;
			runs = 1;
		}
		System.out.println("Run-mode: " + runMode.name().toLowerCase());
		
		/*
		Parse world type from crebe config. Default: STANDARD_WITH_FOOD.
		*/
		if (crebeCfg.has("world_type")) {
			String cfgEntry = crebeCfg.get("world_type");
			for (WorldType wt : WorldType.values()) {
				if (wt.name().equalsIgnoreCase(cfgEntry))
					worldType = wt;
			}
			if (worldType == null) throw new RuntimeException("Unknown world_type \"" + cfgEntry + "\".");
		} else {
			worldType = WorldType.FOOD_WORLD;
		}
		if (runMode == RunMode.PRESENTATION) worldType = WorldType.PRESENTATION;
		System.out.println("World type: " + worldType.toString());
		
		/*
		Parse designed creature.
		*/
		designedCreatureFileName = argParser.getString("-designed");
		System.out.println("Designed creature: " + (designedCreatureFileName == null ? "none" : designedCreatureFileName));
		
		System.out.println();
		
		/*
		Seed rng.
		*/
		if (argParser.isPresent("-seed")) {
			checkSeedMultithreadWarning();
			
			Object seedObj;
			try {
				seedObj = argParser.getInt("-seed");
			} catch (NumberFormatException e) {
				seedObj = argParser.getString("-seed");
			}
			
			int seed = Objects.hashCode(seedObj);
			seedRandom(seed);
		} else {
			seedRandom(); // set seed here as argument when debugging 
		}
		
		
		/*
		Parse entity mutation factors.
		*/
		Arm.ArmWrapper.parseMfConfig(crebeCfg);
		Body.BodyWrapper.parseMfConfig(crebeCfg);
		Ear.EarWrapper.parseMfConfig(crebeCfg);
		Eye.EyeWrapper.parseMfConfig(crebeCfg);
		Fork.ForkWrapper.parseMfConfig(crebeCfg);
		Motor.MotorWrapper.parseMfConfig(crebeCfg);
		Mouth.MouthWrapper.parseMfConfig(crebeCfg);
		Nose.NoseWrapper.parseMfConfig(crebeCfg);
		Spike.SpikeWrapper.parseMfConfig(crebeCfg);
		TouchSensor.TouchSensorWrapper.parseMfConfig(crebeCfg);
		
		/*
		Parse other mutation probabilities.
		*/
		TreeGenotype.parseMpConfig(crebeCfg);
		
		/*
		Show gui.
		*/
		if (HEADLESS) {
			System.out.println("Headless operation.");
		} else if (runMode != RunMode.PRESENTATION) {
			try {
				initTheme();
				evolvedGui = new EvolvedListGUI();
				evolvedGui.setLocationRelativeTo(null);
				evolvedGui.setVisible(true);
			} catch (HeadlessException e) {
				System.err.println(e);
				System.err.println("Run application with -headless");
				System.exit(0);
			}
		}
		
		runAsync();
	}
	
	public static void runAsync() {
		if (!HEADLESS) {
			if (currentBestFitness > Double.NEGATIVE_INFINITY) {
				int res = JOptionPane.showConfirmDialog(evolvedGui,
						"Do you want to clear the creature list before starting a new evolution?"
						, "Clear creature list", JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
				if (res == JOptionPane.YES_OPTION) {
					currentBestFitness = Double.NEGATIVE_INFINITY;
					evolvedGui.clearCreatures();
				}
			}
		}
		
		Thread runThread = new Thread(() -> {
			try {
				if (dumpFitness) {
					clearEaResults(runName);
				}
				
				/*
				Configure threads.
				*/
				int availProcs = Runtime.getRuntime().availableProcessors();
				if (creatureCfg.getInt("threads") == 0) creatureCfg.set("threads", String.valueOf(availProcs));
				System.out.println("Threads for creature_ea: " + creatureCfg.get("threads"));
				if (weightsCfg.getInt("threads") == 0) weightsCfg.set("threads", String.valueOf(runMode == RunMode.FULL ? 1 : availProcs));
				System.out.println("Threads for ann_ea: " + weightsCfg.get("threads"));
				if (fuzzyCfg.getInt("threads") == 0) fuzzyCfg.set("threads", String.valueOf(runMode == RunMode.FULL ? 1 : availProcs));
				System.out.println("Threads for fuzzy_ea: " + fuzzyCfg.get("threads"));
				if (neatCfg.getInt("threads") == 0) neatCfg.set("threads", String.valueOf(runMode == RunMode.FULL ? 1 : availProcs));
				System.out.println("Threads for neat_ea: " + neatCfg.get("threads"));
				
				aborted = false;
				
				if (runMode == RunMode.DESIGNED && brainMethod == null) {
					System.out.println("Please specify brain type using -brain <brain_type>");
				} else {
					Timer totalEt = new Timer();

					for (int run = 0; run < runs && !aborted; run++) {
						currentRun = run;
						if (!HEADLESS && runMode != RunMode.PRESENTATION) evolvedGui.resetInstruments();
						if (runs > 1) System.out.println("Run: " + run);
						
						switch (runMode) {
							case DESIGNED:
								switch (brainMethod) {
									case CTRNN:
										start_ctrnn();
										break;

									case FUZZY_LOGIC:
										start_fuzzy();
										break;

									case NEAT:
										start_neat();
										break;
								}
								break;

							case FULL:
								start_full();
								break;

							case NONE:
								break;

							case PRESENTATION:
								start_presentation();
								break;

						}
					}
					System.out.println("Total runtime: " + totalEt.toString());
				}
			} catch (Exception e) {
				System.err.println("An exception occurred in the main run-thread:\n");
				e.printStackTrace(System.err);
				
				if (!HEADLESS) {
					SwingUtilities.invokeLater(() -> {
						JOptionPane.showMessageDialog(evolvedGui, "An exception occurred in the main run-thread:\n" +
								e.getClass().getSimpleName() + ": " + e.getMessage() + (e.getCause() == null ? "" : "\nCaused by:\n" + e.getCause().getClass().getSimpleName() + ": " + e.getCause().getMessage()), "Error", JOptionPane.ERROR_MESSAGE);
					});
				}
			}
		}, "run-thread");
		runThread.start();
	}
	
	
	public static void newBestFound(Creature c) {
		if (HEADLESS) {
			System.out.println("\nNEW BEST FOUND! fitness = " + c.getFitness());
			File file = new File("exported/" + runName + "_" + (int)c.getFitness() + "_" + Integer.toHexString(c.hashCode()));
			
			try {
				CrebeIO.exportCreature(c, file);
			} catch (IOException e) {
				System.err.println("Unable to export creature to file \"" + file.getAbsolutePath() + "\".");
				System.err.println(e);
			}
		} else {
			evolvedGui.addCreatureToList(c);
			evolvedGui.updateMaxFitness(currentBestFitness);
		}
	}
	
	
	public static void test() throws Exception {
		TreeGenotype tg1 = TreeGenotype.readFromTxtFile("test.txt");
		TreeGenotype tg2 = TreeGenotype.readFromTxtFile("test.txt");
		MainFrame frame = createMainFrame();
		Creature c1 = new ANNCreature(tg1, new WeightsGenotype(tg1));
		Creature c2 = new ANNCreature(tg2, new WeightsGenotype(tg2));
		frame.getRenderWorld().addCreature(c1);
		frame.getRenderWorld().addCreature(c2);
		c1.position.x = 500;
		c1.position.y = 500;
		c2.position.x = 550;
		c2.position.y = 550;
		frame.cycleFollow();
	}
	
	
	public static void start_presentation() throws Exception {
		if (designedCreatureFileName == null && PRESENTATION_CREATURE == null) {
			System.err.println("No creature file specified.");
			return;
		}
		
		String[] creatureFiles = PRESENTATION_CREATURE == null ? new String[] { designedCreatureFileName } : PRESENTATION_CREATURE;
		MainFrame frame = createMainFrame();
		
		int count = 0;
		for (String cf : creatureFiles) {
			String[] cfArray = cf.split("\\.");
			TreeGenotype tg = new TreeGenotype();
			if (cfArray[1].equals("txt")) {
				tg = TreeGenotype.readFromTxtFile(cf);
			} 
			else if (cfArray[1].equals("scd")) {
				tg = TreeGenotype.readFromSCDFile(cf);
			}
			DummyCreature dc = new DummyCreature(tg);
			frame.getRenderWorld().addCreature(dc);
			dc.position.x = count++ * 1000;
			dc.position.y = 500;
			dc.rotation = 0;
		}
	}
	
	
	public static void start_ctrnn() throws Exception {
		TreeGenotype tg = TreeGenotype.readFromTxtFile(designedCreatureFileName);
		
		NetworkWeightsEA ea;
		Population pop;
		if (worldType == WorldType.ARENA_1){
			ea = new NetworkWeightsEA(weightsCfg);
			ea.setCurrentTree(tg);
			ea.evolve();
			
			worldType = WorldType.ARENA_2;
			
			pop = ea.getPopulation();
			ea = new NetworkWeightsEA(weightsCfg);
			ea.setCurrentTree(tg);
			
			ea.setPopulation(pop);
			ea.evolve();
			
			worldType = WorldType.ARENA_WORLD;
		}
		
		ea = new NetworkWeightsEA(weightsCfg);
		if (!HEADLESS) evolvedGui.setCurrentEA(ea);
		ea.setCurrentTree(tg);
		
		attachNewBestFitnessListener(ea);
		attachAvgFitnessListener(ea);
		attachStdFitnessListener(ea);
		attachMaxFitnessListener(ea);
		
		Timer timer = new Timer();
		ea.evolve();
		System.out.println(timer);
		
		if (dumpFitness) {
			System.out.println("Writing results...");
			writeEaResults(ea, runName);
		}
	}
	
	
	public static void start_neat() throws Exception {
		TreeGenotype tg = TreeGenotype.readFromTxtFile(designedCreatureFileName);
		
		NeatEA ea;
		Population pop;
		if (worldType == null){
			ea = new NeatEA(neatCfg, tg);
			ea.evolve();
			
			worldType = WorldType.ARENA_2;
			
			pop = ea.getPopulation();
			ea = new NeatEA(neatCfg, tg);
			
			ea.setPopulation(pop);
			ea.evolve();
			
			worldType = WorldType.ARENA_WORLD;
		}
		
		ea = new NeatEA(neatCfg, tg);
		if (!HEADLESS) evolvedGui.setCurrentEA(ea);
		
		attachNewBestFitnessListener(ea);
		attachAvgFitnessListener(ea);
		attachStdFitnessListener(ea);
		attachMaxFitnessListener(ea);
		
		Timer timer = new Timer();
		ea.evolve();
		System.out.println(timer);
		
		if (dumpFitness) {
			System.out.println("Writing results...");
			writeEaResults(ea, runName);
		}
	}
	
	
	public static void start_neat_test() {
		NeatXOREA ea = new NeatXOREA(neatCfg);
		
		Timer timer = new Timer();
		ea.evolve();
		System.out.println(timer);

	}
	
	
	private static void start_fuzzy() throws Exception {
		TreeGenotype tg = TreeGenotype.readFromTxtFile(designedCreatureFileName);
		
		FuzzyEA ea;
		Population pop;
		if (worldType == WorldType.ARENA_1){
			ea = new FuzzyEA(fuzzyCfg);
			ea.setCurrentTree(tg);
			ea.evolve();
			
			worldType = WorldType.ARENA_2;
			
			pop = ea.getPopulation();
			ea = new FuzzyEA(fuzzyCfg);
			ea.setCurrentTree(tg);
			
			ea.setPopulation(pop);
			ea.evolve();
			
			worldType = WorldType.ARENA_WORLD;
		}
		
		ea = new FuzzyEA(fuzzyCfg);
		if (!HEADLESS) evolvedGui.setCurrentEA(ea);
		ea.setCurrentTree(tg);
		
		attachNewBestFitnessListener(ea);
		attachAvgFitnessListener(ea);
		attachStdFitnessListener(ea);
		attachMaxFitnessListener(ea);
		
		Timer timer = new Timer();
		ea.evolve();
		System.out.println(timer);
		
		if (dumpFitness) {
			System.out.println("Writing results...");
			writeEaResults(ea, runName);
		}
	}
	
	
	private static void start_full() throws Exception {
		CreatureEvolveEA ea = new CreatureEvolveEA(creatureCfg);
		if (!HEADLESS) evolvedGui.setCurrentEA(ea);
		
		attachNewBestFitnessListener(ea);
		attachAvgFitnessListener(ea);
		attachStdFitnessListener(ea);
		attachMaxFitnessListener(ea);
		
		Timer timer = new Timer();
		ea.evolve();
		System.out.println(timer);
		
		if (dumpFitness) {
			System.out.println("Writing results...");
			writeEaResults(ea, runName);
		}
	}
	
	
	public static void attachNewBestFitnessListener(EA ea) {
		ea.addListener(new OnChangeListener() {
			final EA thisNea = ea;

			@Override
			public void handle(OnChangedEvent e) {
				if (e == OnChangedEvent.NEW_BEST_FITNESS) {
					double fitness = thisNea.getWinner().getFitness();

					synchronized (CURRENT_BEST_LOCK) {
						if (fitness > currentBestFitness) {
							currentBestFitness = fitness;
							newBestFound(((Creature) thisNea.getWinner()).copy());
						}
					}
				}
			}
		});
	}
	
	/**
	 * Use attachNewGenerationListener(.) instead.
	 * @param ea
	 * @deprecated
	 */
	@Deprecated
	public static void attachGenerationOverListener(EA ea) {
		ea.addListener(new OnChangeListener() {
			final EA thisNea = ea;

			@Override
			public void handle(OnChangedEvent e) {
				if (e == OnChangedEvent.NEW_GENERATION) {
					evolvedGui.addCreatureToList(((Creature) thisNea.getGenerationWinner()).copy());
				}
			}
		});
	}
	
	
	public static void attachNewGenerationListener(EA ea, Consumer<EA> handler) {
		ea.addListener(new OnChangeListener() {
			final EA thisEa = ea;
			
			@Override
			public void handle(OnChangedEvent e) {
				if (e == OnChangedEvent.NEW_GENERATION) handler.accept(thisEa);
			}
		});
	}
	
	
	public static void attachAvgFitnessListener(EA ea) {
		attachNewGenerationListener(ea, e -> {
			if (!HEADLESS) {
				evolvedGui.updateAvgFitness(e.getAvgFitness());
			}
		});
	}
	
	
	public static void attachStdFitnessListener(EA ea) {
		attachNewGenerationListener(ea, e -> {
			if (!HEADLESS) {
				evolvedGui.updateStdFitness(e.getStdFitness());
			}
		});
	}
	
	
	public static void attachMaxFitnessListener(EA ea) {
		attachNewGenerationListener(ea, e -> {
			if (!HEADLESS) {
				evolvedGui.updateMaxFitness(e.getMaxFitness());
			}
		});
	}


	public static MainFrame createMainFrame() throws HeadlessException {
		World world;
		
		switch (CreativeBehaviour.worldType){
			case STANDARD:
				world = WorldFactory.makeStandardWorld();
				break;
			case FOOD_WORLD:
				world = WorldFactory.makeFoodWorld();
				break;
			case POISON_WORLD:
				world = WorldFactory.makeFoodAndPoisonWorld();
				break;
			case ARENA_1:
				world = WorldFactory.makeArenaTrainingWorld(CreativeBehaviour.worldType);
				break;
			case ARENA_2:
				world = WorldFactory.makeArenaTrainingWorld(CreativeBehaviour.worldType);
				break;
			case ARENA_WORLD:
				world = WorldFactory.makeArenaTrainingWorld(CreativeBehaviour.worldType);
				break;
			case PRESENTATION:
				world = WorldFactory.makePresentationWorld();
				break;
			case FOOD_WORLD_EDGE_SPAWN:
				world = WorldFactory.makeFoodWorldEdgeSpawn();
				break;
			default:
				throw new AssertionError(CreativeBehaviour.worldType.name());	
		}
		
		MainFrame frame = new MainFrame(1600, 900, world);
		frame.setLocationRelativeTo(null);
		frame.startRenderThread();
		frame.setVisible(true);
		return frame;
	}
	
	
	public static void seedRandom(long seed) {
		checkSeedMultithreadWarning();
		
		System.out.println("Seed: " + seed);
		Utils.seed(seed);
		eval.util.Utils.seed(seed);
		anune.util.Utils.seed(seed);
	}
	
	
	public static void seedRandom() {
		long seed = new Random().nextLong();
		System.out.println("Seed: " + seed);
		Utils.seed(seed);
		eval.util.Utils.seed(seed);
		anune.util.Utils.seed(seed);
	}
	
	
	private static void checkSeedMultithreadWarning() {
		if (brainMethod == null) return;
		if (
			(runMode == RunMode.FULL && creatureCfg.getInt("threads") != 1) ||
			(brainMethod == BrainEA.Method.CTRNN && weightsCfg.getInt("threads") != 1) ||
			(brainMethod == BrainEA.Method.FUZZY_LOGIC && fuzzyCfg.getInt("threads") != 1) ||
			(brainMethod == BrainEA.Method.NEAT && neatCfg.getInt("threads") != 1)
		) {
			System.err.println("WARNING! Seeding does not work while multithreaded! Set number of threads in config file(s) to 1, or set random seed.");
			System.err.println("Program will continue in 3 seconds...");
			try {
				Thread.sleep(3000);
			} catch (InterruptedException e) {
			}
		}
	}
	
	
	private static void writeEaResults(EA ea, String runName) throws IOException {
		EAResults res = ea.getResults();
		PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(runName + "_fitness_maximums.txt", true)));
		
		for (Double d : res.fitness_maximums) {
			pw.printf(Locale.ENGLISH, "%.2f", d);
			pw.print("\t");
		}
		pw.println();
		pw.flush();
		pw.close();
		
		pw = new PrintWriter(new BufferedWriter(new FileWriter(runName + "_fitness_averages.txt", true)));
		for (Double d : res.fitness_averages) {
			pw.printf(Locale.ENGLISH, "%.2f", d);
			pw.print("\t");
		}
		pw.println();
		pw.flush();
		pw.close();
		
		pw = new PrintWriter(new BufferedWriter(new FileWriter(runName + "_fitness_stds.txt", true)));
		for (Double d : res.fitness_deviations) {
			pw.printf(Locale.ENGLISH, "%.2f", d);
			pw.print("\t");
		}
		pw.println();
		pw.flush();
		pw.close();
	}
	
	
	private static void clearEaResults(String runName) throws IOException {
		PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(runName + "_fitness_maximums.txt", false)));
		pw.flush();
		pw.close();
		pw = new PrintWriter(new BufferedWriter(new FileWriter(runName + "_fitness_averages.txt", false)));
		pw.flush();
		pw.close();
		pw = new PrintWriter(new BufferedWriter(new FileWriter(runName + "_fitness_stds.txt", false)));
		pw.flush();
		pw.close();
	}
}