package creativebehaviour.ents;

import creativebehaviour.ents.misc.OutputEntity;
import creativebehaviour.ents.misc.AttachPoint;
import creativebehaviour.ents.misc.EntType;
import creativebehaviour.ents.wrapper.Wrapper;
import anune.ANN;
import creativebehaviour.graphics.Render;
import creativebehaviour.util.Pointf;
import creativebehaviour.util.Utils;
import eval.util.Config;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Line2D;
import java.awt.geom.NoninvertibleTransformException;

/**
 *
 * @author Bjørnar W. Alvestad
 */
public class Arm extends Entity implements OutputEntity {

	public static class ArmWrapper extends Wrapper {

		public float length = 50;  // pixels?
		public float speed = 90;  // deg/sec?
		public float strength = 0.3f; // Ratio between 0..1?
		public float joint_angle = 60;  // Degrees
		
		private static float length_mf;
		private static float speed_mf;
		private static float strength_mf;
		private static float joint_angle_mf;
		
		public static void parseMfConfig(Config cfg) {
			length_mf = cfg.getFloat("arm_length_mf");
			speed_mf = cfg.getFloat("arm_speed_mf");
			strength_mf = cfg.getFloat("arm_strength_mf");
			joint_angle_mf = cfg.getFloat("arm_jointangle_mf");
		}

		public ArmWrapper() {
			super(ATTACH_POINTS, EntType.ARM);
		}

		@Override
		public void mutate(double mutateFactor) {
			length += mutateFactor * length_mf;
			if (length < 2) {
				length = 2;
			}
			
			speed += mutateFactor * speed_mf;
			if (speed < 2) {
				speed = 2;
			}
			
			strength += mutateFactor * strength_mf;
			if (strength < 0) {
				strength = 0;
			}
			
			joint_angle += mutateFactor * joint_angle_mf;
			if (joint_angle < 0) {
				joint_angle = 0;
			} else if (joint_angle > 120) {
				joint_angle = 120;
			}
		}

		@Override
		public Entity createEntity() {
			return new Arm(this);
		}

		@Override
		public boolean isSensor() {
			return false;
		}

		@Override
		public boolean isActuator() {
			return true;
		}

		@Override
		public Wrapper deepCopy() {
			ArmWrapper a = new ArmWrapper();
			a.joint_angle = joint_angle;
			a.length = length;
			a.speed = speed;
			a.strength = strength;
			return a;
		}

		@Override
		public double calcUpkeep() {
			double sum = 0;
			sum += 3 * length / 50.0;
			sum += speed / 90.0;
			sum += strength / 0.3;
			sum += 2 * joint_angle / 45.0;
			return sum / 70.0;
		}

		@Override
		public EntType getEntType() {
			return EntType.ARM;
		}
		
	}

	public static final double BASE_UPKEEP = 1.0;
	public static final int ATTACH_POINTS = 1;
	private static final BasicStroke ARM_THICKNESS_STROKE = new BasicStroke(12, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);

	public ArmWrapper wrapper;
	private final double jointAngleRad;
	private final double speedRad;
	
	
	public Arm() {
		this(new ArmWrapper());
	}

	
	public Arm(ArmWrapper w) {
		super(EntType.ARM, BASE_UPKEEP);
		
		wrapper = w;
		jointAngleRad = Math.toRadians(wrapper.joint_angle);
		rotation = (Utils.randomDouble()-0.5)*2*jointAngleRad;
		speedRad = Math.toRadians(wrapper.speed);

		addAttachPoint(0, 0, 0);

		updateAttachPointOrigins();
	}
	

	public double getLength() {
		return wrapper.length;
	}

	
	public void setLength(double length) {
		wrapper.length = (float) length;
		updateAttachPointOrigins();
	}

	
	public void setArmAttachment(Entity ent) {
		assert attachPoints.size() == 1;
		attachPoints.get(0).setChild(ent);
	}

	
	private void updateAttachPointOrigins() {
		assert attachPoints.size() == 1;
		Pointf p = attachPoints.get(0).parentRelativePos;
		p.x = wrapper.length;
	}

	
	private int dir = Utils.randomEvent(0.5) ? 1 : -1;
	@Override
	public void render(Graphics2D g) throws NoninvertibleTransformException {
//		rotation += dir * speedRad * 0.0222;
//		
//		if (rotation < -jointAngleRad || rotation > jointAngleRad) {
//			rotation = jointAngleRad * dir;
//			dir *= -1;
//		}
		
		resetTransform();
		transform.rotate(rotation);
		applyTransformation(g, transform);

		Render.renderOutlinedLine(new Line2D.Double(0, 0, wrapper.length, 0), Color.BLACK, getColor(), Render.STANDARD_STROKE, ARM_THICKNESS_STROKE, g);

		g.setColor(Color.BLACK);
		AttachPoint ap = attachPoints.get(0);
		ap.render(g);

		revertTransformation(g, transform);
		super.render(g);
	}

	
	@Override
	public double performAction(double value) {
		this.rotation += (value - 0.5) * wrapper.speed * 0.001;
		if (rotation < -jointAngleRad) rotation = -jointAngleRad+0.03;
		else if (rotation > jointAngleRad) rotation = jointAngleRad-0.03;
		return value;
	}
	
	
	@Override
	public double performActions(double[] activation) {
		return performAction(activation[0]);
	}
	
}
