package creativebehaviour.ents;

import creativebehaviour.ents.misc.OutputEntity;
import creativebehaviour.ents.misc.AttachPoint;
import creativebehaviour.ents.misc.EntType;
import creativebehaviour.ents.misc.InputEntity;
import creativebehaviour.ents.wrapper.Wrapper;
import creativebehaviour.graphics.Render;
import creativebehaviour.models.Creature;
import creativebehaviour.simulation.World;
import creativebehaviour.util.Pointf;
import eval.util.Config;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.NoninvertibleTransformException;

/**
 *
 * @author Bjørnar W. Alvestad
 */
public class Body extends Entity implements OutputEntity, InputEntity {

	
	public static class BodyWrapper extends Wrapper {
		public float radius = 40;
		public float rotationSpeed = 0.04f;

		private static float radius_mf;
		private static float rotationSpeed_mf;
		
		public BodyWrapper() {
			super(ATTACH_POINT_ANGLES.length, EntType.BODY);
		}
		
		public static void parseMfConfig(Config cfg) {
			radius_mf = cfg.getFloat("body_radius_mf");
			rotationSpeed_mf = cfg.getFloat("body_rot_speed_mf");
		}

		@Override
		public void mutate(double mutateFactor) {
			radius += mutateFactor * radius_mf;
			rotationSpeed += mutateFactor * rotationSpeed_mf;
			
			if (radius < 30f) {
				radius = 30f;
			}
			if (rotationSpeed < 0.08f) {
				rotationSpeed = 0.08f;
			}
		}

		@Override
		public Entity createEntity() {
			throw new RuntimeException("Unable to create body entity. Use createBody() instead.");
		}
		
		public Body createBody(Creature bodyOf) {
			return new Body(bodyOf);
		}
		
		@Override
		public boolean isSensor() {
			return true;
		}
		
		@Override
		public boolean isActuator() {
			return true;
		}

		@Override
		public Wrapper deepCopy() {
			BodyWrapper b = new BodyWrapper();
			b.radius = radius;
			b.rotationSpeed = rotationSpeed;
			return b;
		}

		@Override
		public double calcUpkeep() {
			double sum = 0.0;
			sum += radius / 40.0;
			sum += rotationSpeed / 0.1;
			return sum / 2.0;
		}

		@Override
		public EntType getEntType() {
			return EntType.BODY;
		}
	}
	

	public static final double BASE_UPKEEP = 1.0;
	public static final double[] ATTACH_POINT_ANGLES = { 0, 45, 90, 135, 180, 225, 270, 315 }; // Degrees
	private static final Stroke STROKE = new BasicStroke(2);
	private final Creature bodyOf;

	
	
	public BodyWrapper wrapper;
	
	
	public Body(Creature bodyOf) {
		this(new BodyWrapper(), bodyOf);
		
	}
	
	
	public Body(BodyWrapper w, Creature bodyOf) {
		super(EntType.BODY, BASE_UPKEEP);
		
		wrapper = w;
		
		for (int i = 0; i < ATTACH_POINT_ANGLES.length; i++) {
			addAttachPoint(0, 0, Math.toRadians(ATTACH_POINT_ANGLES[i]));
		}
		
		updateAttachPointOrigins();
		
		this.bodyOf = bodyOf;
	}
	
	
	public Creature getBodyOf() {
		return bodyOf;
	}
	
	
	public void setRadius(float radius) {
		wrapper.radius = radius;
		updateAttachPointOrigins();
	}
	
	
	public float getRadius() {
		return wrapper.radius;
	}
	
	
	public void setBodyAttachment(Entity ent, int index) {
		assert attachPoints.size() == ATTACH_POINT_ANGLES.length;
		
		if (index >= ATTACH_POINT_ANGLES.length) throw new IllegalArgumentException("Invalid Body attachment index.");
		if (ent == this) throw new IllegalArgumentException("Cant add attachment to itself.");
		
		attachPoints.get(index).setChild(ent);
	}
	
	
	private void updateAttachPointOrigins() {
		for (AttachPoint ap : attachPoints) {
			//ap.parentRelativePos.x = wrapper.radius;
			ap.parentRelativePos.x = wrapper.radius * Math.cos(ap.getRotation());
			ap.parentRelativePos.y = wrapper.radius * Math.sin(ap.getRotation());
		}
	}
	
	
	@Override
	public void render(Graphics2D g) throws NoninvertibleTransformException {
		resetTransform();
		Shape shape = new Ellipse2D.Double(-wrapper.radius, -wrapper.radius, wrapper.radius*2, wrapper.radius*2);
		
		// Render body
		Render.renderOutlinedShape(shape, Color.BLACK, getColor(), STROKE, g);
		g.setColor(Color.BLACK);
		
		// Render children
		int i = 0;
		for (AttachPoint ap : attachPoints) {
			ap.render(g);
		}
		
		super.render(g);
	}

	
	@Override
	protected void globalPosRecursive(AffineTransform t) {
		if (bodyOf != null) {
			t.translate(bodyOf.position.x, bodyOf.position.y);
			t.rotate(bodyOf.rotation);
		}
	}
	
	
	/**
	 * Rotation action.
	 * @param value 
	 */
	@Override
	public double performAction(double value) {
		value -= 0.5;
		bodyOf.rotation += value * wrapper.rotationSpeed;
		return Math.abs(value);
	}
	
	
	/**
	 * Get energy level.
	 * @param world
	 * @return 
	 */
	@Override
	public double getInputValue(World world) {
		return this.bodyOf.getEnergy() / Creature.MAX_ENERGY;
	}
	
	
	@Override
	public double performActions(double[] activation) {
		// Moving
		double value = activation[0];
		double forceFactor = value * 1; // TODO: hardcoded body movement strength
		double angle = getGlobalRotation();
		
		Pointf force = new Pointf(Math.cos(angle), Math.sin(angle));
		force.normalize();
		force.scale(-forceFactor);
		bodyOf.applyForce(force);
		
		// Turning
		double turn = (activation[1] - 0.5) * 2;
		bodyOf.rotation += wrapper.rotationSpeed * turn;
		value += Math.abs(turn*0.5);
		
		return value;
	}

	
	@Override
	public double[] getInputData(World world) {
		return new double[] { getInputValue(world) };
	}
	
}
