package creativebehaviour.ents;

import creativebehaviour.ents.misc.OutputEntity;
import creativebehaviour.ents.misc.AttachPoint;
import creativebehaviour.ents.misc.Renderable;
import creativebehaviour.ents.misc.EntType;
import creativebehaviour.ents.misc.InputEntity;
import creativebehaviour.models.Creature;
import creativebehaviour.util.Pointf;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.io.Serializable;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.util.ArrayList;
import java.util.Iterator;


public abstract class Entity implements Serializable, Iterable<Entity>, Renderable {
	
	public static final boolean RENDER_GLOBAL_POS = false;
	
	protected transient AffineTransform transform = new AffineTransform();
	protected static final AffineTransform IDENTITY_TRANSFORM = new AffineTransform();
	
	protected double upkeep;
	public final EntType entType;
	protected final ArrayList<AttachPoint> attachPoints;
	public AttachPoint attachedTo;
	protected double rotation = 0.0;
	private Creature rootCreature;

	
	/**
	 * 
	 * @param type
	 * @param upkeep 
	 * @param attachedTo
	 */
	public Entity(EntType type, double upkeep, AttachPoint attachedTo) {
		this.upkeep = upkeep;
		this.attachPoints = new ArrayList<>();
		this.entType = type;
		this.attachedTo = attachedTo;
	}
	
	/**
	 * 
	 * @param type
	 * @param upkeep 
	 */
	public Entity(EntType type, double upkeep) {
		this(type, upkeep, null);
	}
	
	
	public Color getColor() {
		return rootCreature == null ? Color.WHITE : rootCreature.treeGenotype.color;
	}
	
	
	public void setRotation(double angle) {
		this.rotation = angle;
	}
	
	
	public double getRotation() {
		return this.rotation;
	}
	
	
	public double getUpkeep() {
		return upkeep;
	}
	
	
	/**
	 * Adds an attachment point
	 * @param x
	 * @param y
	 * @param child
	 * @param rotation
	 * @return 
	 */
	protected AttachPoint addAttachPoint(double x, double y, Entity child, double rotation) {
		AttachPoint a = new AttachPoint(this, x, y, child, rotation);
		attachPoints.add(a);
		return a;
	}
	
	
	protected final AttachPoint addAttachPoint(double x, double y, double rotation) {
		return addAttachPoint(x, y, null, rotation);
	}
	
	
	public ArrayList<AttachPoint> getAttachPoints() {
		return attachPoints;
	}
	
	
	public Creature getParentCreature() {
		if (attachedTo != null && attachedTo.parent != null) {
			return attachedTo.parent.getParentCreature();
		} else if (entType == EntType.BODY) {
			return ((Body)this).getBodyOf();
		}
		return null;
	}
	
	
	/**
	 * Add the specified entity as a child to this. Skips skipNulls number of attachment points before assigning the child.
	 * @param child
	 * @param skipNulls
	 */
	public void addChildEntity(Entity child, int skipNulls) {
		for (AttachPoint ap : attachPoints) {
			if (ap.getChild() == null) {
				if (skipNulls == 0) {
					ap.setChild(child);
					child.rootCreature = rootCreature;
					return;
				}
				else skipNulls--;
			}
		}
		
		throw new RuntimeException(String.format("No available attachment points on %s.", getClass().getSimpleName()));
	}
	
	
	protected void resetTransform() {
		transform.setToIdentity();
	}
	
	
	protected static void revertTransformation(Graphics2D g, AffineTransform transform) throws NoninvertibleTransformException {
		transform.invert();
		g.transform(transform);
	}
	
	
	protected static void applyTransformation(Graphics2D g, AffineTransform transform) {
		g.transform(transform);
	}
	
	
	protected void renderAttachPoints(Graphics2D g) throws NoninvertibleTransformException {
		for (AttachPoint ap : attachPoints) {
			ap.render(g);
		}
	}

	@Override
	public void render(Graphics2D g) throws NoninvertibleTransformException {
		if (RENDER_GLOBAL_POS) {
			g.setColor(Color.RED);
			g.fillOval(-2, -2, 4, 4);
		}
	}
	
	
	@Override
	public Iterator<Entity> iterator() {
		return new EntIterator();
	}
	
	
	public void forEachAttachPoint(AttachPointConsumer consumer) {
		attachPoints.forEach(AP -> {
			consumer.accept(AP, AP.getChild());
		});
	}
	
	
	public Pointf getGlobalPos() {
		AffineTransform t = new AffineTransform();
		globalPosRecursive(t);
		return new Pointf(t.getTranslateX(), t.getTranslateY());
	}
	
	
	protected void globalPosRecursive(AffineTransform t) {
		if (attachedTo != null) {
			attachedTo.parent.globalPosRecursive(t);
			t.translate(attachedTo.parentRelativePos.x, attachedTo.parentRelativePos.y);
			t.rotate(rotation + attachedTo.getRotation());
		}
		
	}
	
	
	public double getGlobalRotation() {
		double rot = rotation;
		
		if (attachedTo != null) {
			rot += attachedTo.getRotation();
			if (attachedTo.parent != null) {
				rot += attachedTo.parent.getGlobalRotation();
			}
		} else if (entType == EntType.BODY) {
			rot += ((Body)this).getBodyOf().rotation;
		}
		
		while (rot > 2 * Math.PI) rot -= 2 * Math.PI;
		while (rot < 0) rot += 2 * Math.PI;
		
		return rot;
	}
	
	
	public Creature getRootCreature() {
		return rootCreature;
	}

	
	public void setRootCreature(Creature rootCreature) {
		this.rootCreature = rootCreature;
	}
	
	
	// -------------- MISC ENTITY STUFF --------------
	
	
	@Retention(RetentionPolicy.RUNTIME)
	@Target(ElementType.FIELD)
	public @interface GenotypeField {
	}


	/**
	 * Iterate over all children.
	 */
	public class EntIterator implements Iterator<Entity> {
		private int count = 0;
		
		@Override
		public boolean hasNext() {
			return count < attachPoints.size() && attachPoints.get(count).getChild() != null;
		}

		@Override
		public Entity next() {
			return attachPoints.get(count++).getChild();
		}
	}
	
	
	@FunctionalInterface
	public static interface AttachPointConsumer {
		public void accept(AttachPoint ap, Entity child);
	}
	
	public boolean entityIsSensor() {
		return this instanceof InputEntity;
	}
	
	public boolean entityIsActuator() {
		return this instanceof OutputEntity;
	}
	
	
}
