package creativebehaviour.ents;

import creativebehaviour.ents.misc.DebugColors;
import creativebehaviour.ents.misc.EntType;
import creativebehaviour.ents.misc.InputEntity;
import creativebehaviour.CreativeBehaviour;
import static creativebehaviour.ents.misc.DebugColors.DEBUG_STROKE;
import creativebehaviour.ents.wrapper.Wrapper;
import creativebehaviour.graphics.Render;
import creativebehaviour.simulation.Food;
import creativebehaviour.simulation.FoodWorld;
import creativebehaviour.simulation.World;
import creativebehaviour.simulation.WorldEntity;
import creativebehaviour.util.Pointf;
import creativebehaviour.util.Utils;
import eval.util.Config;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.Ellipse2D;
import java.awt.geom.NoninvertibleTransformException;

/**
 *
 * @author Bjørnar W. Alvestad
 */
public class Eye extends Entity implements InputEntity {
	

	public static class EyeWrapper extends Wrapper {
		public float FOV = 25; // Degrees
		public float range = 500;
		public boolean can_measure_dist = false;
		
		private static float fov_mf;
		private static float range_mf;
		private static float meas_dist_mf;

		public EyeWrapper() {
			super(EntType.EYE);
		}
		
		public static void parseMfConfig(Config cfg) {
			fov_mf = cfg.getFloat("eye_fov_mf");
			range_mf = cfg.getFloat("eye_range_mf");
			meas_dist_mf = cfg.getFloat("eye_measdist_mf");
		}

		@Override
		public void mutate(double mutateFactor) {
			FOV += mutateFactor * fov_mf;
			range += mutateFactor * range_mf;
			if (!can_measure_dist) can_measure_dist = Utils.randomEvent(meas_dist_mf);
			if (FOV < 0) FOV = 0;
			else if (FOV > 180) FOV = 180;
			if (range < 0) range = 0;
		}

		@Override
		public Entity createEntity() {
			return new Eye(this);
		}
		
		@Override
		public boolean isSensor() {
			return true;
		}
		
		@Override
		public boolean isActuator() {
			return false;
		}

		@Override
		public Wrapper deepCopy() {
			EyeWrapper e = new EyeWrapper();
			e.FOV = FOV;
			e.range = range;
			e.can_measure_dist = can_measure_dist;
			return e;
		}

		@Override
		public double calcUpkeep() {
			double sum = 0.0;
			sum += FOV / 5.0;
			sum += range / 100.0;
			sum *= can_measure_dist ? 2 : 1;
			return sum / 2.0;
		}

		@Override
		public EntType getEntType() {
			return EntType.EYE;
		}
	}

	public static final double BASE_UPKEEP = 1.0;
	
	private static final int RADIUS = 12;
	private static final int IRIS_RADIUS = 6;
	private static final int IRIS_DISPLACEMENT = 1;
	private static final int PUPIL_RADIUS = 3;
	private static final Color IRIS_COLOR = new Color(60, 150, 255);
	
	public EyeWrapper wrapper;
	
	private final Shape eyeShape;
	private final Shape irisShape;
	private final Shape pupilShape;
	
	private final double fovRad;
	

	public Eye() {
		this(new EyeWrapper());
	}
	
	
	public Eye(EyeWrapper w) {
		super(EntType.EYE, BASE_UPKEEP);
		
		wrapper = w;
		fovRad = Math.toRadians(wrapper.FOV);
		
		eyeShape = new Ellipse2D.Double(-RADIUS, -RADIUS, RADIUS*2, RADIUS*2);
		irisShape = new Ellipse2D.Double(RADIUS-IRIS_RADIUS*2+IRIS_DISPLACEMENT, -IRIS_RADIUS, IRIS_RADIUS*2, IRIS_RADIUS*2);
		pupilShape = new Ellipse2D.Double(RADIUS-PUPIL_RADIUS*2+IRIS_DISPLACEMENT, -PUPIL_RADIUS, PUPIL_RADIUS*2, PUPIL_RADIUS*2);
	}
	
	
	@Override
	public void render(Graphics2D g) throws NoninvertibleTransformException {
		Render.renderOutlinedShape(eyeShape, Color.BLACK, Color.WHITE, Render.STANDARD_STROKE, g);
		Render.renderOutlinedShape(irisShape, Color.BLACK, CreativeBehaviour.DEBUG && seeFood ? IRIS_COLOR.brighter().brighter() : IRIS_COLOR, Render.STANDARD_STROKE, g);
		g.setColor(Color.BLACK);
		g.fill(pupilShape);
		
		// Drawing the FOV
		if (CreativeBehaviour.DEBUG) {
			g.setColor(DebugColors.EYE_DEBUG_COLOR);
			g.setStroke(DEBUG_STROKE);
			float dX = this.wrapper.range * (float) Math.cos(Math.toRadians(this.wrapper.FOV/2.0));
			float dY = this.wrapper.range * (float) Math.sin(Math.toRadians(this.wrapper.FOV/2.0));
			g.drawLine(0, 0, (int) dX, (int) dY);
			g.drawLine((int) dX, (int) dY, (int) dX, (int) -dY);
			g.drawLine((int) dX, (int) -dY, 0, 0);
			g.setColor(Color.BLACK);
		}
		
		super.render(g);
	}
	
	
	private boolean seeFood = false;
	@Override
	public double getInputValue(World world) {
		double vision = 0;
		
		Pointf eyePos = this.getGlobalPos();
		double eyeRot = getGlobalRotation();
		
		if (world instanceof FoodWorld) {
			FoodWorld fworld = (FoodWorld)world;
			for (Food f : fworld.getFood()) {
				if (Math.abs(eyePos.distanceTo(f.position)) < wrapper.range) {
					double dX = (f.position.x - eyePos.x);
					double dY = (f.position.y - eyePos.y);
					double foodAngle = Math.atan2(dY, dX);

					foodAngle -= eyeRot;

					while (foodAngle < -Math.PI) foodAngle += 2 * Math.PI;
					while (foodAngle > Math.PI) foodAngle -= 2 * Math.PI;

					if (foodAngle <= fovRad * 0.5 && foodAngle >= -fovRad * 0.5) {
						vision += 1;
						seeFood = true;
					}
				}
			}
		}
		
		return vision;
	}
	
	
	/**
	 * Get normalized distance and direction to closest object.
	 * @param world
	 * @return {distance, direction}
	 */
	@Override
	public double[] getInputData(World world) {
		double[] data = { 1.1, 0.0 };
		
		Pointf eyePos = this.getGlobalPos();
		double eyeRot = getGlobalRotation();
		double closestDist = Double.POSITIVE_INFINITY;
		
		for (WorldEntity e : world) {
			Pointf ePos = e.getPos();
			double distToObj = eyePos.distanceTo(ePos);

			if (distToObj < wrapper.range) {
				double dX = (ePos.x - eyePos.x);
				double dY = (ePos.y - eyePos.y);
				double eAngle = Math.atan2(dY, dX);

				eAngle -= eyeRot;

				while (eAngle < -Math.PI) eAngle += 2 * Math.PI;
				while (eAngle > Math.PI) eAngle -= 2 * Math.PI;

				if (eAngle <= fovRad * 0.5 && eAngle >= -fovRad * 0.5) { // is inside fov
					if (distToObj < closestDist) {
						closestDist = distToObj;
						data[0] = distToObj / wrapper.range;
						data[1] = eAngle * 2 / fovRad;
					}
				}
			}
		}
		
		seeFood = Double.isFinite(closestDist);
		
		return data;
	}
	
	
	
}
