package creativebehaviour.graphics;

import creativebehaviour.CreativeBehaviour;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import javax.swing.JPanel;
import creativebehaviour.models.Creature;
import creativebehaviour.simulation.World;
import creativebehaviour.simulation.WorldFactory;
import creativebehaviour.simulation.WorldType;
import java.awt.Color;
import java.awt.Font;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;

/**
 *
 * @author Bjørnar W. Alvestad
 */
public class RenderPanel extends JPanel {
	private static final Font debugFont = new Font("Consolas", Font.BOLD, 14);
	
	public World world;
	
	private double zoom = 1.0;
	private boolean paused = false;
	private volatile int follow = -1;
	
	private boolean mouseDown = false;
	private Point mouseDownPoint = new Point();
	private Point topLeft = new Point();
	private Point centerWhenDown = new Point();
	
	private int WIDTH;
	private int HEIGHT;
	
	
	public RenderPanel(int width, int height) {
		switch (CreativeBehaviour.worldType){
			case STANDARD:
				world = WorldFactory.makeStandardWorld();
				break;
			case FOOD_WORLD:
				world = WorldFactory.makeFoodWorld();
				break;
			case POISON_WORLD:
				world = WorldFactory.makeFoodAndPoisonWorld();
				break;
			case ARENA_1:
				world = WorldFactory.makeArenaTrainingWorld(CreativeBehaviour.worldType);
				break;
			case ARENA_2:
				world = WorldFactory.makeArenaTrainingWorld(CreativeBehaviour.worldType);
				break;
			case ARENA_WORLD:
				world = WorldFactory.makeArenaTrainingWorld(CreativeBehaviour.worldType);
				break;
			case MULTI_AGENT:
				throw new UnsupportedOperationException("Multi agent world");
			default:
				throw new AssertionError(CreativeBehaviour.worldType.name());	
		}
	}

	public RenderPanel(int width, int height, World world) {
		this.world = world;
		this.WIDTH = width;
		this.HEIGHT = height;
		this.topLeft = new Point(-WIDTH/2, -HEIGHT/2);
		
		this.follow = world.getNumCreatures() - 1;
		
		setPreferredSize(new Dimension(width, height));

		addComponentListener(new ComponentListener() {
			@Override
			public void componentResized(ComponentEvent e) {
				WIDTH = e.getComponent().getWidth();
				HEIGHT = e.getComponent().getHeight();
				topLeft = new Point(-WIDTH/2, -HEIGHT/2);
			}

			@Override
			public void componentMoved(ComponentEvent e) {
			}

			@Override
			public void componentShown(ComponentEvent e) {
			}

			@Override
			public void componentHidden(ComponentEvent e) {
			}
		});
		
		addMouseWheelListener((MouseWheelEvent e) -> {
			zoom += e.getWheelRotation() * -(0.1 * zoom);
		});
		
		addMouseListener(new MouseListener() {
			@Override
			public void mouseClicked(MouseEvent e) {
			}

			@Override
			public void mousePressed(MouseEvent e) {
				if (e.getButton() == 1) {
					mouseDownPoint = e.getPoint();
					centerWhenDown = new Point(topLeft);
					mouseDown = true;
				}
				else if (e.getButton() == 3) {
//					world.spawnFood((e.getPoint().x + (-topLeft.x) - WIDTH / 2), (e.getPoint().y + (-topLeft.y) - HEIGHT / 2));
				}
			}

			@Override
			public void mouseReleased(MouseEvent e) {
				if (e.getButton() == 1)
					mouseDown = false;
			}

			@Override
			public void mouseEntered(MouseEvent e) {
			}

			@Override
			public void mouseExited(MouseEvent e) {
			}
		});
		
		addMouseMotionListener(new MouseMotionListener() {
			@Override
			public void mouseDragged(MouseEvent e) {
				
				if (mouseDown) {
					int mx = e.getX();
					int my = e.getY();
					
					int mdx = mx - mouseDownPoint.x;
					int mdy = my - mouseDownPoint.y;
					
					topLeft.x = (int)(centerWhenDown.x + mdx / zoom);
					topLeft.y = (int)(centerWhenDown.y + mdy / zoom);
				}
				
			}

			@Override
			public void mouseMoved(MouseEvent e) {
			}
		});	
	}
	
	
	/**
	 * Render this panel.
	 */
	public void render() {
		this.repaint();
	}
	
	
	public void tickWorld() {
		if (!paused)
			world.tick();
	}
	
	
	public void togglePause() {
		paused = !paused;
	}
	
	
	/*public void addCreatureToRenderList(Creature c) {
		synchronized (worldEnts) {
			c.position.x = (worldEnts.size() % 10) * 500;
			c.position.y = (worldEnts.size() / 10) * 500;
			worldEnts.add(c);
		}
	}*/
	
	
	@Override
	public void paint(Graphics g) {
		Graphics2D g2 = (Graphics2D) g;
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		super.paint(g);
		
		try {
			if (follow >= 0) {
				Creature c = world.getCreature(follow);
				topLeft.x = (int)-c.position.x;
				topLeft.y = (int)-c.position.y;
				
				if (CreativeBehaviour.DEBUG) {
					g.setFont(debugFont);
					g.drawString("  " + c.sensorLegendStr + "      " + c.actuatorLegendStr, 5, debugFont.getSize());
					g.drawString(c.getSensorBufferStr() + " -> " + c.getActuatorStr(), 5, debugFont.getSize()*2);
				}
			}
			if (CreativeBehaviour.worldType == WorldType.ARENA_1 || CreativeBehaviour.worldType == WorldType.ARENA_2 || CreativeBehaviour.worldType == WorldType.ARENA_WORLD) {
				boolean red = true;
				for (Creature c : world.creatures) {
					if (red) {
						g.setColor(Color.DARK_GRAY);
						g.fillRect(0+WIDTH/20, 0+HEIGHT/20, WIDTH/2-WIDTH/10, HEIGHT/20);
						g.setColor(Color.RED);
						// 100% hp = WIDTH/2-WIDTH/10
						int x = WIDTH/20+5;
						int y = HEIGHT/20+5;
						int width = (int) Math.floor((WIDTH/2-WIDTH/10) * (c.getEnergy() / Creature.MAX_ENERGY));
						int height = HEIGHT/20;
						g.fillRect(x, y, width, height); // TODO scale with hp
						red = false;
					}
					else {
						g.setColor(Color.DARK_GRAY);
						g.fillRect(WIDTH/2+WIDTH/20, 0+HEIGHT/20, WIDTH/2-WIDTH/10, HEIGHT/20);
						g.setColor(Color.BLUE);
						int y = HEIGHT/20+5;
						int width = (int) Math.floor((WIDTH/2-WIDTH/10) * (c.getEnergy() / Creature.MAX_ENERGY));
						int height = HEIGHT/20;
						int x = (WIDTH/2+WIDTH/20-5) + (WIDTH/2-WIDTH/10 - width);
						g.fillRect(x, y, width, height); // TODO scale with hp
					}
				}
			}
			
			g2.translate(topLeft.x + (WIDTH >> 1), topLeft.y + (HEIGHT >> 1));
			g2.translate(-topLeft.x, -topLeft.y);
			g2.scale(zoom, zoom);
			g2.translate(topLeft.x, topLeft.y);
			
			if (world != null) {
				world.render(g2);
			}
			

		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		
		
	}

	
	public void cycleFollow() {
		if (follow == world.getNumCreatures() - 1) {
			follow = -1;
		} else follow++;
	}
	
}
