package creativebehaviour.models;

import creativebehaviour.models.physical.TreeGenotype;
import creativebehaviour.CreativeBehaviour;
import creativebehaviour.ents.misc.AttachPoint;
import creativebehaviour.ents.Body;
import creativebehaviour.ents.Entity;
import creativebehaviour.ents.misc.InputEntity;
import creativebehaviour.ents.wrapper.NullWrapper;
import creativebehaviour.ents.misc.OutputEntity;
import creativebehaviour.ents.misc.Renderable;
import creativebehaviour.ents.wrapper.Wrapper;
import creativebehaviour.fuzzy.EntityVariable;
import creativebehaviour.fuzzy.Variable;
import creativebehaviour.simulation.Food;
import creativebehaviour.simulation.Simulation;
import creativebehaviour.simulation.World;
import creativebehaviour.simulation.WorldEntity;
import creativebehaviour.simulation.WorldType;
import creativebehaviour.util.tree.Node;
import creativebehaviour.util.Pointf;
import creativebehaviour.util.Utils;
import eval.models.Genotype;
import eval.models.Phenotype;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.util.ArrayList;
import java.util.Locale;
import java.util.Stack;

/**
 *
 * @author Bjørnar W. Alvestad
 */
public abstract class Creature extends Phenotype implements WorldEntity {
	
	/** 0.0 = max friction, 1.0 = no friction. */
	public static final double MOVEMENT_FRICTION = 0.8;
	/** Upkeep each tick that is always present. */
	public static final double STANDARD_UPKEEP = 0.5;
	
	public static final int SIMULATION_TICKS = CreativeBehaviour.crebeCfg.getInt("simulation_ticks");
	public static final int MAX_ENERGY = CreativeBehaviour.crebeCfg.getInt("start_energy");
	public static final int FITNESS_RUNS = CreativeBehaviour.crebeCfg.getInt("fitness_eval_runs");
	
	/** Creature position. */
	public final Pointf position;
	
	/** Creature velocity. */
	public final Pointf velocity;
	
	/** Creature rotation, in radians. */
	public double rotation;
	
	/** Genotype that describes the physical layout of the creature. */
	public final TreeGenotype treeGenotype;
	
	/** The root entity of this creature. */
	public Body rootEnt;
	
	/** Number of ticks lived during the last simulation. */
	protected int ticksLived;
	
	private int foodEaten;
	
	/** Current energy level for this creature. */
	protected double energy;
	
	/** Indicator if creature has just taken damage **/
	public boolean invoulnerable = false;
	
	/** Reset counter for invoulnerable tag **/
	protected int invCounter = 30;
	
	/** Energy cost for current actions */
	private double performActionCost;
	
	/** Total creature upkeep. */
	public final double upkeep;
	
	/** Input entities in this creature. */
	public final ArrayList<InputEntity> inputEntities;
	
	/** Output entities in this creature. */
	public final ArrayList<OutputEntity> outputEntities;
	
	/** Buffer used to store values from input entities. */
	private final float[] sensorBuffer;
	
	/** Buffer used to store values to output entities. */
	private final float[] actuatorBuffer;
	
	protected final EntityVariable[] sensorLegend;
	
	protected final EntityVariable[] actuatorLegend;
	
	public final String sensorLegendStr;
	
	public final String actuatorLegendStr;
	
	/** Buffer for multiple fitness runs. */
	private final double[] fitnessRunsBuffer;
	
	private final double[] entityActivationBuffer;
	
	/** The world type (scenario) this creature is evolved for. */
	public WorldType worldType;
	
	
	/**
	 * Create a new creature, using the supplied TreeGenotype.
	 * @param treeGenotype The genotype that defines the physical entity layout of the creature.
	 * @param brainGenotype Custom genotype associated with the action inference system used in this creature.
	 */
	protected Creature(TreeGenotype treeGenotype, Genotype brainGenotype) {
		super(brainGenotype);
		
		this.position       = new Pointf();
		this.velocity       = new Pointf();
		this.inputEntities  = new ArrayList<>();
		this.outputEntities = new ArrayList<>();
		this.treeGenotype   = treeGenotype;
		this.rotation       = Utils.randomDouble(0, 2*Math.PI);
		this.energy         = MAX_ENERGY;
		this.worldType      = CreativeBehaviour.worldType;
		
		parseTree(treeGenotype);
		getIOEntitiesFromTree();
		
		
		int buffSize = 0;
		for (InputEntity inp : inputEntities) {
			buffSize += ((Entity)inp).entType.inputLegend.length;
		}
		this.sensorBuffer = new float[buffSize];
		
		buffSize = 0;
		int activationBuffMaxLength = 0;
		for (OutputEntity outp : outputEntities) {
			int len = ((Entity)outp).entType.outputLegend.length;
			buffSize += len;
			if (len > activationBuffMaxLength) activationBuffMaxLength = len;
		}
		this.actuatorBuffer = new float[buffSize];
		this.entityActivationBuffer = new double[activationBuffMaxLength];
		this.fitnessRunsBuffer = new double[FITNESS_RUNS];
		
		
		this.sensorLegend = new EntityVariable[sensorBuffer.length];
		int count = 0;
		for (InputEntity inp : inputEntities) {
			Entity e = (Entity)inp;
			for (Variable v : e.entType.inputLegend) {
				sensorLegend[count] = new EntityVariable(e, v, count);
				count++;
			}
		}
		
		this.actuatorLegend = new EntityVariable[actuatorBuffer.length];
		count = 0;
		for (OutputEntity out : outputEntities) {
			Entity e = (Entity)out;
			for (Variable v : e.entType.outputLegend) {
				actuatorLegend[count] = new EntityVariable(e, v, count);
				count++;
			}
		}
		
		// Calculate the total upkeep
		double upkeepTmp = 0;
		for (Wrapper w : treeGenotype.tree) upkeepTmp += w.calcUpkeep();
		this.upkeep = upkeepTmp;
		
		
		StringBuilder strBuilder = new StringBuilder();
		String[] strings = new String[sensorLegend.length];
		for (int i = 0; i < sensorLegend.length; i++) {
			strBuilder.append("  %-6s");
			strings[i] = sensorLegend[i].getVar().shortName;
		}
		
		sensorLegendStr = String.format(Locale.ENGLISH, strBuilder.substring(2), (Object[])strings);	
		
		strBuilder = new StringBuilder();
		strings = new String[actuatorLegend.length];
		for (int i = 0; i < actuatorLegend.length; i++) {
			strBuilder.append("  %-6s");
			strings[i] = actuatorLegend[i].getVar().shortName;
		}
		
		actuatorLegendStr = String.format(Locale.ENGLISH, strBuilder.substring(2), (Object[])strings);
		
		//System.out.println(Arrays.toString(sensorLegend) + "\t" + Arrays.toString(actuatorLegend));
	}
	
	
	protected final void setEqual(Creature original) {
		this.fitness = original.fitness;
		this.position.setEqual(original.position);
		this.velocity.setEqual(original.velocity);
		this.rotation = original.rotation;
		this.energy = original.energy;
		this.ticksLived = original.ticksLived;
		this.foodEaten = original.foodEaten;
		this.worldType = original.worldType;
	}
	
	
	public abstract Creature copy();
	
	
	/**
	 * Traverse TreeGenotype in order to create the entity instances of this creature.
	 * (convert genotype into phenotype).<br><br>
	 * This involves converting wrapper objects into entity objects.
	 * @param treeg 
	 */
	private void parseTree(TreeGenotype treeg) {
		if (!(treeg.tree.root.obj instanceof Body.BodyWrapper))
			throw new RuntimeException("Root entity must be body.");
		
		rootEnt = new Body((Body.BodyWrapper)treeg.tree.root.obj, this);
		rootEnt.setRootCreature(this);
		parseTreeRecursive(treeg.tree.root, rootEnt);
	}
	
	
	/**
	 * Recursive helper method for the parseTree(.) method.
	 * @param rootNode
	 * @param rootEnt 
	 */
	private void parseTreeRecursive(Node<Wrapper> rootNode, Entity rootEnt) {
		int skipNullCounter = 0;
		
		for (Node<Wrapper> childNode : rootNode) {
			if (childNode.obj == null || childNode.obj instanceof NullWrapper) {
				skipNullCounter++;
			} else {
				Entity childEnt = childNode.obj.createEntity();
				rootEnt.addChildEntity(childEnt, skipNullCounter);
				parseTreeRecursive(childNode, childEnt);
			}
		}
	}
	
	
	/**
	 * Finds input and output entities in this creature, and places them in
	 * the inputEntities list and outputEntities list respectively.
	 */
	private void getIOEntitiesFromTree() {
		Stack<Entity> stack = new Stack<>();
		stack.push(this.rootEnt);
		
		while (!stack.isEmpty()) {
			Entity entity = stack.pop();
			
			// Careful fakking with this. List order is important to fuzzy stuff
			if (entity.entityIsSensor()) this.inputEntities.add((InputEntity)entity);
			if (entity.entityIsActuator()) this.outputEntities.add((OutputEntity)entity);
			
			for (AttachPoint ap : entity.getAttachPoints()) {
				Entity child = ap.getChild();
				if (child != null) stack.push(child);
			}
		}
	}
	
	
	/**
	 * Calculate fitness for this creature.
	 * @return 
	 */
	@Override
	public double calculateFitness() {
		for (int i = 0; i < FITNESS_RUNS; i++) {
			Simulation sim = new Simulation(CreativeBehaviour.worldType, SIMULATION_TICKS);
			sim.addCreature(this);
			sim.runSimulation();
			fitnessRunsBuffer[i] = this.fitness; // relies on World setting this.fitness at simulation end
		}
		
		return Utils.avg(fitnessRunsBuffer);
	}

	
	@Override
	public void render(Graphics2D g) throws NoninvertibleTransformException {
		AffineTransform origTransform = g.getTransform();
		g.translate(position.x, position.y);
		g.rotate(rotation);
		
		if (rootEnt != null) {
			rootEnt.render(g);
			if (CreativeBehaviour.runMode != CreativeBehaviour.RunMode.PRESENTATION) {
				g.setColor(Color.BLACK);
				g.drawString(String.format(Locale.ENGLISH, "%.1f", energy), -20, 0);
				if (!isAlive()) g.drawString("DEAD", -20, 10);
			}
		}
		
		g.setTransform(origTransform);
	}
	
	
	@Override
	public void tick(World world) {
		if (isAlive()) {
			int counter = 0;
			for (InputEntity input : inputEntities) {
				double[] entInput = input.getInputData(world);
				for (int i = 0; i < entInput.length; i++) {
					sensorBuffer[counter++] = (float)entInput[i];
				}
			}
			
			customCreatureTick(sensorBuffer, actuatorBuffer);
			
			counter = 0;
			performActionCost = 0.0;
			for (OutputEntity output : outputEntities) {
				int num = ((Entity)output).entType.outputLegend.length;
				for (int i = 0; i < num; i++) {
					entityActivationBuffer[i] = actuatorBuffer[counter++];
				}
				performActionCost += Math.abs(output.performActions(entityActivationBuffer));
			}
			
			ticksLived++;

			position.add(velocity);
			velocity.scale(MOVEMENT_FRICTION);
			
			
			if (world.WRAP_AROUND) {
				if (position.x > world.WIDTH) position.x = 0;
				if (position.x < 0) position.x = world.WIDTH;

				if (position.y > world.HEIGHT) position.y = 0;
				if (position.y < 0) position.y = world.HEIGHT;
			}
			else {
				if (position.x > world.WIDTH) position.x = world.WIDTH;
				if (position.x < 0) position.x = 0;

				if (position.y > world.HEIGHT) position.y = world.HEIGHT;
				if (position.y < 0) position.y = 0;
			}
			
			if (world.ENERGY_DRAIN) {
				energy -= STANDARD_UPKEEP + performActionCost;
				if (energy < 0) energy = 0;
			}
			
			if (invoulnerable) {
				invCounter--;
			}
			if (invCounter <= 0) {
				invCounter = 30;
				invoulnerable = false;
			}
		}
	}
	
	
	protected abstract void customCreatureTick(float[] sensorData, float[] actuatorBuffer);
	
	
	public String getSensorBufferStr() {
		return Utils.toString(sensorBuffer);
	}
	
	
	public String getActuatorStr() {
		return Utils.toString(actuatorBuffer);
	}
	
	
	public void setFitness(double fitness) {
		this.fitness = fitness;
	}
	
	
	public boolean isAlive() {
		return energy > 0;
	}
	
	
	public int getTicksLived() {
		return ticksLived;
	}
	
	
	public void resetTicksLived() {
		ticksLived = 0;
	}
	
	
	public double getEnergy() {
		return energy;
	}
	
	public int getFoodEaten(){
		return this.foodEaten;
	}
	
	
	public void addEnergy(double energy) {
		this.energy += energy;
		if (this.energy > MAX_ENERGY) this.energy = MAX_ENERGY;
	}
	
	
	public void eatFood(Food f) {
		addEnergy(f.nourishment);
		if (f.nourishment > 0) {
			foodEaten++;
		}
		else {
			foodEaten--;
		}
		f.nourishment = 0;
	}
	
	public void takeDamage(int damage) {
		if (!this.invoulnerable) {
			this.energy -= damage;
			this.invoulnerable = true;
		}
	}
	
	
	public void resetEnergy() {
		energy = MAX_ENERGY;
	}
	
	
	public double getUpkeep() {
		return upkeep;
	}
	
	
	public void applyForce(Pointf force) {
		velocity.add(force);
	}
	
	
	public void resurrect() {
		ticksLived = 0;
		foodEaten = 0;
		energy = MAX_ENERGY;
		velocity.x = 0;
		velocity.y = 0;
	}
	
	
	public TreeGenotype getTreeGenotype() {
		return treeGenotype;
	}

	
	@Override
	public Pointf getPos() {
		return position;
	}

	
	@Override
	public Type getWorldEntType() {
		return Type.CREATURE;
	}
	
}
