package creativebehaviour.models.ann;

import anune.ANN;
import creativebehaviour.CreativeBehaviour;
import creativebehaviour.ents.wrapper.Wrapper;
import creativebehaviour.util.tree.Tree;
import static creativebehaviour.models.ann.WeightsGenotype.*;

/**
 *
 * @author Endre
 */
public abstract class NetworkFactory {
	
	public static final int[] ANN_HIDDEN_CONFIGURATION;
	
	static {
		String cfgStr = CreativeBehaviour.crebeCfg.get("ann_hidden_config");
		String[] hiddenCfgStr = cfgStr.split(",");
		ANN_HIDDEN_CONFIGURATION = new int[hiddenCfgStr.length];
		
		for (int i = 0; i < hiddenCfgStr.length; i++) {
			ANN_HIDDEN_CONFIGURATION[i] = Integer.parseInt(hiddenCfgStr[i]);
		}
	}
	
	
	/**
	 * Creates ANN using the supplied tree and the default hidden layer configuration
	 * specified in the configuration file.
	 * @param t
	 * @return 
	 */
	public static ANN makeANN(Tree<Wrapper> t) {
		return NetworkFactory.makeANN(t, ANN_HIDDEN_CONFIGURATION);
	}
	
	
	/**
	 * Creates ANN using the supplied tree and hidden configuration.
	 * @param t
	 * @param hiddenConfig
	 * @return 
	 */
	public static ANN makeANN(Tree<Wrapper> t, int[] hiddenConfig) {
		int[] annConfig = new int[hiddenConfig.length + 2];
		System.arraycopy(hiddenConfig, 0, annConfig, 1, hiddenConfig.length);
		
		final int lastIndex = annConfig.length - 1;
		for (Wrapper entwrapper : t) {
			if (entwrapper.isSensor()) annConfig[0] += entwrapper.entType.getInputLenght();
			if (entwrapper.isActuator()) annConfig[lastIndex] += entwrapper.entType.getOutputLenght();
		}
		
		return new ANN(annConfig);
	}
	
	
	/**
	 * Creates ANN using the supplied tree and weights. Hidden layer config pulled from
	 * configuration file.
	 * @param t
	 * @param weights
	 * @return 
	 */
	public static ANN makeANN(Tree<Wrapper> t, float[][] weights) {
		ANN network = NetworkFactory.makeANN(t);
		applyWeightsToAnn(network, weights);
		return network;
	}
	
	
	/**
	 * Creates ANN using tree, hidden config and weights.
	 * @param t
	 * @param hiddenConfig
	 * @param weights
	 * @return 
	 */
	public static ANN makeANN(Tree<Wrapper> t, int[] hiddenConfig, float[][] weights) {
		ANN network = NetworkFactory.makeANN(t, hiddenConfig);
		applyWeightsToAnn(network, weights);
		return network;
	}
	
	
	/**
	 * Creates ANN using the supplied configuration and weights.
	 * @param configuration
	 * @param weights
	 * @return 
	 */
	public static ANN makeANN(int[] configuration, float[][] weights) {
		ANN network = makeANN(configuration);
		applyWeightsToAnn(network, weights);
		return network;
	}
	
	
	/**
	 * Creates ANN using the supplied configuration.
	 * @param configuration
	 * @return 
	 */
	public static ANN makeANN(int[] configuration) {
		return new ANN(configuration);
	}
	
	
	// --- PRIVATE ---
	
	private static void applyWeightsToAnn(ANN ann, float[][] weights) {
		ann.updateWeights(weights[WEIGHTS]);
		ann.updateTaus(weights[TAUS]);
		ann.updateGains(weights[GAINS]);
		ann.updateBias(weights[BIASES]);
	}
	
}
