package creativebehaviour.simulation;

import creativebehaviour.ents.misc.Renderable;
import creativebehaviour.models.Creature;
import creativebehaviour.util.Utils;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.NoninvertibleTransformException;
import java.util.ArrayList;
import java.util.Iterator;

/**
 *
 * @author Endrus
 */
public abstract class World implements Renderable, Iterable<WorldEntity> {
	
	/** List containing only the creatures in this world. */
	public final ArrayList<Creature> creatures; // creatures is a subset of worldEnts, with just the creatures.
	
	protected final Object WORLD_ENTS_LOCK = new Object(); // TODO: maybe LOCK can be removed.
	
	public final int WIDTH;
	public final int HEIGHT;
	
	public final boolean WRAP_AROUND;
	public final boolean ENERGY_DRAIN;
	
	private int tickCounter = 0;
	public final WorldType worldType;
	
	
	public World(int width, int height, WorldType worldType) {
		this(width, height, worldType, true, true);
	}
	
	public World(int width, int height, WorldType worldType, boolean wrapAround, boolean energyDrain) {
		this.WIDTH = width;
		this.HEIGHT = height;
		this.worldType = worldType;
		this.creatures = new ArrayList<>();
		
		this.WRAP_AROUND = wrapAround;
		this.ENERGY_DRAIN = energyDrain;
	}
	
	
	public void tick() {
		for (WorldEntity e : this) {
			e.tick(this);
		}
		customWorldTick(tickCounter);
		tickCounter++;
	}
	
	
	protected abstract void customWorldTick(int tickCounter);
	protected abstract void customWorldRender(Graphics2D g) throws NoninvertibleTransformException;
	
	
	/**
	 * Thread safe.
	 * @param c 
	 * @param spawnAreaSize 
	 */
	public void addCreature(Creature c, double spawnAreaSize) {
		c.resurrect();
		
		c.position.x = WIDTH * 0.25 * spawnAreaSize * Utils.randomDouble(-1, 1) + WIDTH * 0.5;
		c.position.y = HEIGHT * 0.25 * spawnAreaSize * Utils.randomDouble(-1, 1) + HEIGHT * 0.5;
		
//		c.position.x = WIDTH*0.5*(Utils.randomDouble()+0.5)*spawnAreaSize;
//		c.position.y = HEIGHT*0.5*(Utils.randomDouble()+0.5)*spawnAreaSize;
		
		c.rotation = Utils.randomDouble() * 2 * Math.PI;
		
		synchronized (WORLD_ENTS_LOCK) {
			creatures.add(c);
		}
	}
	
	
	public void addCreature(Creature c) {
		addCreature(c, 1);
	}
	
	
	@Override
	public void render(Graphics2D g) throws NoninvertibleTransformException {
		g.setColor(Color.BLACK);
		g.drawRect(0, 0, WIDTH, HEIGHT);
		
		synchronized (WORLD_ENTS_LOCK) {
			for (Creature c : creatures) {
				c.render(g);
			}
		}
		
		customWorldRender(g);
	}
	
	
	public boolean hasAliveCreatures() {
		for (int i = 0; i < creatures.size(); i++) {
			if (creatures.get(i).isAlive()) return true;
		}
		return false;
	}
	

	public boolean continueCheck() {
		return hasAliveCreatures();
	}
	
	
	public boolean finished() {
		return false;
	}
	
	
	public abstract void setFitnessOnCreatures();
	
	
	public int getTickCount() {
		return tickCounter;
	}
	
	
	/**
	 * Creates a shallow copy of the creature list.
	 * @return 
	 */
	public ArrayList<Creature> getCreatures() {
		synchronized (WORLD_ENTS_LOCK) {
			return new ArrayList<>(creatures);
		}
	}
	
	
	/**
	 * @return The number of entities in this world, including creatures.
	 */
	public int getNumOfEnts() {
		return creatures.size() + getNumCustomEnts();
	}
	
	
	/**
	 * @return The number of creatures in this world.
	 */
	public int getNumCreatures() {
		return creatures.size();
	}
	
	
	public Creature getCreature(int index) {
		return creatures.get(index);
	}

	
	@Override
	public Iterator<WorldEntity> iterator() {
		return new WorldEntIterator(customWorldCollectionIterator());
	}
	
	
	/**
	 * Override this in subclasses in order to incude custom world entities in
	 * world iterator. Return null if this world subclass has no custom entity
	 * collection.
	 * @return 
	 */
	protected abstract Iterator customWorldCollectionIterator();
	
	/**
	 * Gets the number of custom entities in this subclass. Return 0 if
	 * this world subclass has no custom entities.
	 * @return 
	 */
	protected abstract int getNumCustomEnts();
	
	
	
	private class WorldEntIterator implements Iterator<WorldEntity> {
		
		private final Iterator customIt;
		private Iterator currentIt;
		
		
		WorldEntIterator(Iterator customIt) {
			this.customIt = customIt;
			currentIt = creatures.iterator();
		}
		
		
		@Override
		public boolean hasNext() {
			return currentIt == null ? false : currentIt.hasNext();
		}
		
		
		@Override
		public WorldEntity next() {
			WorldEntity ent = (WorldEntity) currentIt.next();
			if (!currentIt.hasNext()) currentIt = customIt;
			return ent;
		}

		
		@Override
		public void remove() {
			currentIt.remove();
		}
		
	}
	
	
}
