package creativebehaviour.util.io;

import creativebehaviour.ents.misc.EntType;
import creativebehaviour.ents.Entity;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;

/**
 *
 * @author Bjørnar W. Alvestad
 */
public class EntityCoding {

	public static byte[] encodeEntity(Entity ent) {
		// id derived from entity type enum
		int entId = ent.entType.ordinal();
		
		// Set up output streams
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		BufferedOutputStream bos = new BufferedOutputStream(baos);
		ObjectOutputStream oos = null;
		
		try {
			oos = new ObjectOutputStream(bos);
		} catch (IOException ex) {
			throw new EntityEncodeException("Failed to create output stream.", ex);
		}
		
		try {
			// Write entId to first byte
			oos.writeByte(entId);
		} catch (IOException ex) {
			throw new EntityEncodeException(String.format("Error while writing ent_id (%d) to byte array.", entId), ex);
		}
		
		try {
			// Iterate fields from ent class
			for (Field field : ent.entType.entClass.getFields()) {
				if (isGenotypeField(field)) {
					Class<?> fieldType = field.getType();
					
					// Write field data to stream
					if (fieldType.isPrimitive()) {
						
						if (fieldType == float.class)
						oos.writeFloat(field.getFloat(ent));

						else if (fieldType == boolean.class)
							oos.writeBoolean(field.getBoolean(ent));

						else if (fieldType == double.class)
							oos.writeDouble(field.getDouble(ent));

						else if (fieldType == int.class)
							oos.writeInt(field.getInt(ent));

						else if (fieldType == byte.class)
							oos.writeByte(field.getByte(ent));

						else if (fieldType == short.class)
							oos.writeShort(field.getShort(ent));

						else if (fieldType == long.class)
							oos.writeLong(field.getLong(ent));

						else if (fieldType == char.class)
							oos.writeChar(field.getChar(ent));
						
					} else if (isSerializable(fieldType)) {
						// Field is object
						oos.writeObject(field.get(ent));
					}
					
					else {
						throw new EntityEncodeException(
								String.format("Invalid field type: \"%s\" in entity %s. Datatype %s not supported. Class must implement Serializable.",
								String.format("%s %s %s", Modifier.toString(field.getModifiers()), fieldType.getSimpleName(), field.getName()),
								ent.getClass().getSimpleName(),
								fieldType.getSimpleName()
								));
					}
					
				}
			}
		} catch (SecurityException | IllegalArgumentException | IllegalAccessException | IOException e) {
			throw new EntityEncodeException("Error while writing entity fields to byte array.", e);
		}
		
		try {
			oos.flush();
		} catch (IOException ex) {
			throw new EntityEncodeException("Error while flushing output stream.", ex);
		}
		
		return baos.toByteArray();
	}
	
	
	public static Entity decodeEntity(byte[] bytes) {
		// Set up input streams
		ByteArrayInputStream bais = new ByteArrayInputStream(bytes);
		BufferedInputStream bis = new BufferedInputStream(bais);
		ObjectInputStream ois = null;
		
		try {
			ois = new ObjectInputStream(bis);
		} catch (IOException ex) {
			throw new EntityEncodeException("Failed to create input stream.", ex);
		}

		int entId = -1;
		try {
			// First byte denotes entity type
			entId = ois.readByte();
		} catch (IOException ex) {
			throw new EntityDecodeException("Unable to read ent_id from byte array.", ex);
		}
		
		if (entId >= EntType.values().length) {
			throw new EntityDecodeException(String.format("ent_id outside range. Expected <= %d, got %d.", EntType.values().length-1, entId));
		}
		EntType entType = EntType.values()[entId];

		// Create new instance of appropriate entity type.
		Entity ent = null;
		try {
			ent = entType.entClass.newInstance();
		} catch (InstantiationException | IllegalAccessException ex) {
			throw new EntityDecodeException(String.format("Unable to create new instance of entity type %s.", entType.name()), ex);
		}

		try {
			// Iterate fields in the respective entity-type class
			for (Field field : entType.entClass.getFields()) {
				if (isGenotypeField(field)) {
					Class<?> fieldType = field.getType();

					// Read data from stream, write to entity obj.
					if (fieldType.isPrimitive()) {
						
						if (fieldType == float.class)
							field.setFloat(ent, ois.readFloat());

						else if (fieldType == boolean.class)
							field.setBoolean(ent, ois.readBoolean());

						else if (fieldType == double.class)
							field.setDouble(ent, ois.readDouble());

						else if (fieldType == int.class)
							field.setInt(ent, ois.readInt());

						else if (fieldType == byte.class)
							field.setByte(ent, ois.readByte());

						else if (fieldType == short.class)
							field.setShort(ent, ois.readShort());

						else if (fieldType == long.class)
							field.setLong(ent, ois.readLong());

						else if (fieldType == char.class)
							field.setChar(ent, ois.readChar());
						
					} else if (isSerializable(fieldType)) {
						// Field is object
						try {
							field.set(ent, ois.readObject());
						} catch (IOException | ClassNotFoundException e) {
							throw new EntityDecodeException(String.format("Error while trying to decoding object field: %s.", field.toString()), e);
						}
						
					}

					else {
						throw new EntityDecodeException(
							String.format("Invalid field type: \"%s\" in entity %s. Datatype %s not supported. Class must implement Serializable.",
							String.format("%s %s %s", Modifier.toString(field.getModifiers()), fieldType.getSimpleName(), field.getName()),
							ent.getClass().getSimpleName(),
							fieldType.getSimpleName()
							));
					}
				}
			}
		} catch (IOException | IllegalAccessException e) {
			throw new EntityDecodeException("Error while reading entity fields from byte array.", e);
		} catch (ArrayIndexOutOfBoundsException e) {
			throw new EntityDecodeException(String.format("Not enough bytes in the array to successfully decode entity: %s.", entType.name()), e);
		}

		// If bytes remain in buffer, hassard.
		if (bais.available() != 0)
			throw new EntityDecodeException(String.format("Unknown error. %d byte(s) remain in byffer.", bais.available()));

		return ent;
	}
	
	
	/**
	 * Helper method to check a field for the GenotypeField annotation.
	 * @param f
	 * @return true if the field represented by the supplied Field argument has the GenotypeField annotation. false otherwise.
	 */
	private static boolean isGenotypeField(Field f) {
		return f.getAnnotation(Entity.GenotypeField.class) != null;
	}
	
	
	/**
	 * Helper method to check if a class implements the Serializable interface.
	 * @param clazz
	 * @return true if the class represented by the supplied Class argument implements the Serializable interface. false otherwise. 
	 */
	private static boolean isSerializable(Class clazz) {
		return Serializable.class.isAssignableFrom(clazz);
	}
	
	
//	public static int encodedLength(Class clazz) {
//		int length = 0;
//		
//		for (Field field : clazz.getFields()) {
//			if (isGenotypeField(field)) {
//				
//			}
//		}
//	}
	
	
	/**
	 * Encode runtime exception.
	 */
	public static class EntityEncodeException extends RuntimeException {
		public EntityEncodeException(String message, Exception cause) {
			super(message, cause);
		}

		public EntityEncodeException(String message) {
			super(message);
		}
	}
	
	
	/**
	 * Decode runtime exception.
	 */
	public static class EntityDecodeException extends RuntimeException {
		public EntityDecodeException(String message, Exception cause) {
			super(message, cause);
		}

		public EntityDecodeException(String message) {
			super(message);
		}
	}
	
}
