package eval.ea;

import eval.models.Crossover;
import eval.models.Genotype;
import eval.models.Phenotype;
import eval.models.Population;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import eval.ea.selection.adult.AdultSelection;
import eval.ea.selection.adult.GenerationalReplacementAdultSelection;
import eval.ea.selection.parent.ParentSelection;
import eval.util.Config;
import eval.util.Pair;
import eval.util.Utils;



public abstract class EA implements Serializable {
	
	// config stuff
	protected final Config cfg;
	protected final int      threads;
	protected final boolean  dynamicworkload;
	protected final boolean  verbose;
	protected final boolean  quiet;
	protected final double   fitnessLimit;
	protected final int      generationLimit;
	protected final int      elites;
	protected final double   crossoverrate;
	protected final double   mutationrate;
	protected final int      populationSize;
	protected final double   overprodFactor;
	protected final boolean  recalculateFitness;
	protected final int      numOffsprings;
	protected final boolean  minimizeFitness;
	
	protected final Crossover       crossoverType;
	protected final ParentSelection parentSelection;
	protected final AdultSelection  adultSelection;
	
	// runtime stuff
	protected boolean    running;
	protected boolean    paused;
	protected final Object pauseObj;
	protected boolean    aborted;
	protected int        generation;
	protected double     bestfitness;
	protected double     avgfitness;
	protected double     stdfitness;
	protected Phenotype  winner;
	protected Phenotype  genwinner;
	protected int        brokerindex;
	
	protected Population           population;
	protected Population           offspring;
	protected ArrayList<Phenotype> elitesBuffer;
//	private ArrayList<Phenotype> adultSelectionBuffer;
	
	private transient ArrayList<OnChangeListener> listeners;
	protected final ArrayList<Double> fitnessAverages;
	protected final ArrayList<Double> fitnessBests;
	protected final ArrayList<Double> fitnessDeviations;
	
	
	public EA(Config cfg) {
		this.cfg = cfg;
		
		parentSelection    = ParentSelection.parseSelection(cfg);
		adultSelection     = AdultSelection.parseSelection(cfg);
		crossoverType      = Crossover.parseCrossover(cfg);
		
		// required config
		populationSize     = cfg.getInt("population_size");
		overprodFactor     = adultSelection instanceof GenerationalReplacementAdultSelection ? 1.0 : cfg.getDouble("overprod_factor");
		fitnessLimit       = cfg.getDouble("target_fitness");
		generationLimit    = cfg.getInt("max_generations");
		crossoverrate      = cfg.getDouble("crossover_rate");
		mutationrate       = cfg.getDouble("mutation_rate");

		// optional config
		threads            = cfg.getInt("threads", 1);
		dynamicworkload    = cfg.getBoolean("dynamic_workload", false);
		quiet              = cfg.getBoolean("quiet", false);
		verbose            = cfg.getBoolean("verbose", false) && !quiet;
		elites             = cfg.getInt("elites", 0);
		recalculateFitness = cfg.getBoolean("recalculate_fitness", false);
		minimizeFitness    = cfg.getBoolean("minimize_fitness", false);
		
		listeners          = new ArrayList<>();
		fitnessAverages    = new ArrayList<>();
		fitnessBests       = new ArrayList<>();
		fitnessDeviations  = new ArrayList<>();
		
		if (overprodFactor < 1.0) {
			throw new RuntimeException("Overproduction factor must be greater than 1.0 (is " + overprodFactor + ").");
		}
		
		elitesBuffer  = new ArrayList<>(elites);
		population    = new Population(0);
		numOffsprings = populationSize;
		offspring     = new Population(numOffsprings);
		pauseObj      = new Object();
		
//		adultSelectionBuffer = new ArrayList<>(numOffsprings);
		
		reset();
	}
	
	
	public abstract Population generateInitalPopulation(int size);
	public abstract Phenotype  generatePhenotype(Genotype g);
	
	
	/**
	 * Reset and make ready for evolve
	 */
	protected void reset() {
		running = false;
		paused = false;
		aborted = false;
		
		winner = null;
		genwinner = null;
		
		population.clear();
		offspring.clear();
		elitesBuffer.clear();
		
		fitnessAverages.clear();
		fitnessDeviations.clear();
		fitnessBests.clear();
		
		generation = -1;
		bestfitness = minimizeFitness ? Double.MAX_VALUE : 0;
		avgfitness = 0;
		stdfitness = 0;
		brokerindex = 0;
	}
	
	
	protected void notifyListeners(OnChangedEvent event) {
		for (OnChangeListener l : listeners) {
			l.handle(event);
		}
	}
	
	
	/**
	 * 
	 * @return true if max fitness is reached
	 */
	public boolean evolve() {
		printVerbose("Starting evolution with " + threads + " thread(s).");
		long gentimer1, gentimer0 = System.currentTimeMillis();
		
		reset();
		running = true;
		
		// Initialization
		population = generateInitalPopulation(populationSize);
		calculateFitnesses(population);
		notifyListeners(OnChangedEvent.FITNESS_CALCULATED);
		
		while (true) {
			brokerindex = 0;
			generation++;
			notifyListeners(OnChangedEvent.NEW_GENERATION);

			// Check paused status
			while (paused) {
				try {
					synchronized (pauseObj) {
						pauseObj.wait();
					}
				} catch (InterruptedException e) {
				}
			}
			
			population.sort(minimizeFitness);

			
			// Fitness data and stats retreival
			if (minimizeFitness)
				genwinner = population.getMinFitness();
			else
				genwinner = population.getMaxFitness();
			if (winner == null || 
					(!minimizeFitness && genwinner.getFitness() > winner.getFitness()) || 
					(minimizeFitness && genwinner.getFitness() < winner.getFitness())) {
				winner = genwinner;
				notifyListeners(OnChangedEvent.NEW_BEST_FITNESS);
			}
			avgfitness = population.getAvgFitness();
			stdfitness = population.getStdFitness();
			bestfitness = genwinner.getFitness();

			fitnessAverages.add(avgfitness);
			fitnessBests.add(bestfitness);
			fitnessDeviations.add(stdfitness);
			
			gentimer1 = System.currentTimeMillis();
			long genelapsed = gentimer1 - gentimer0;
			
			// Print info
			if (!quiet) {
				if (verbose) {
					double progress = (generation / (double)generationLimit) * 100.0;
					System.out.println();
					System.out.printf("G# %d     best_f: %.4f     avg_f: %.4f     SD: %.4f     Progress: %.2f%%     Elapsed: %d ms.\n", generation, bestfitness, avgfitness, stdfitness, progress, genelapsed);
					System.out.println("Generation winner: "+ genwinner);
				} else {
					StringBuilder sb = new StringBuilder();
					sb.append("\r");
					Utils.buildProgressbar(sb, 50, generation, generationLimit);
					sb.append("  Gen: ");
					sb.append(generation);
					System.out.print(sb);
					System.out.flush();
				}
			}
			
			if (isDone()) break;
			gentimer0 = System.currentTimeMillis();
			
			// Elites extraction
			elitesBuffer.clear();
			for (int i = 0; i < elites; i++) {
				elitesBuffer.add(population.get(i));
			}
			
			// Parent selection
			List<Pair<Phenotype>> parents = parentSelection.selectParents(population, numOffsprings, !minimizeFitness); // TODO: optimize, use regular array instead of List<Pair>?
			
			// Reproduction w/ crossover + mutation
			offspring.clear();
			for (Pair<Phenotype> pair : parents) {
				Genotype p1 = pair.first.getGenotype();
				Genotype p2 = pair.second.getGenotype();
				int bestParent = 0;
				if (pair.first.getFitness() > pair.second.getFitness()) bestParent = minimizeFitness ? -1 : 1;
				else if (pair.first.getFitness() < pair.second.getFitness()) bestParent = minimizeFitness ? 1 : -1;
				Genotype offspr;
				if (Utils.randomEvent(crossoverrate)) {
					offspr = p1.crossover(p2, crossoverType, bestParent);
				} else {
					offspr = p1.copy();
				}

				offspr.mutate(mutationrate);

				Phenotype child = generatePhenotype(offspr);
				offspring.add(child);
			}
			
			// Evaluate offspring
			calculateFitnesses(offspring);
			notifyListeners(OnChangedEvent.FITNESS_CALCULATED);
			
			// Re-evaluate elites, if configured
			if (recalculateFitness && !elitesBuffer.isEmpty()) {
				for (Phenotype elite : elitesBuffer) {
					elite.evaluate();
				}
			}
			
			// Adult selection
			adultSelection.selectAdults(offspring, population, populationSize - elites, !minimizeFitness);
			
			// Reintroduce elites
			population.add(elitesBuffer);
//			for (Phenotype p : population) {
//				System.out.println(p.getFitness());
//			}
			assert population.getCount() == populationSize;
		}
		
		running = false;
		return generation < generationLimit;
	}
	

	
	public void addListener(OnChangeListener listener) {
		listeners.add(listener);
	}
	
	
	public void removeListeners() {
		listeners.clear();
	}
	
	
	protected boolean interrupt() {
		return false;
	}
	
	
	/**
	 * True if max fitness or generation limit is reached.
	 * @return 
	 */
	protected boolean isDone() {
		if (aborted) {
			notifyListeners(OnChangedEvent.FITNESS_LIMIT_REACHED);
			printVerbose("ABORTED");
			return true;
		}
		else if (generation >= generationLimit) {
			notifyListeners(OnChangedEvent.GENERATION_LIMIT_REACHED);
			printVerbose("GENERATION LIMIT REACHED");
			return true;
		}
		else if ((minimizeFitness && bestfitness <= fitnessLimit) ||
			(!minimizeFitness && bestfitness >= fitnessLimit)) {
			notifyListeners(OnChangedEvent.FITNESS_LIMIT_REACHED);
			printVerbose("FITNESS LIMIT REACHED");
			return true;
		}
		return !running;
	}
	
	public void setPopulation(Population pop){
		this.population = pop;
	}
	
	public Phenotype getWinner() {
		return winner;
	}
	
	
	public Phenotype getGenerationWinner() {
		return genwinner;
	}
	
	
	public double getMaxFitness() {
		return bestfitness;
	}
	
	
	public double getAvgFitness() {
		return avgfitness;
	}
	
	
	public double getStdFitness() {
		return stdfitness;
	}
	
	
	protected void calculateFitnesses(Population pop) {
		if (threads <= 1) {
			for (Phenotype p : pop) {
				checkAndCalcFitness(p);
			}
		}
		else {
			if (dynamicworkload) 
				dynamicFitnessCalculation(pop);
			else 
				staticFitnessCalculation(pop);
		}
	}
	
	
	private synchronized int getNextThreadIndex() {
		return brokerindex++;
	}
	
	
	private void dynamicFitnessCalculation(Population pop) {
		int total = pop.getCount();
		Thread[] pool = new Thread[threads];
		
		for (int t = 0; t < threads; t++) {
			final int max = total;
			final int id  = t;
			pool[t] = new Thread(() -> {
				while (true) {
					int i = getNextThreadIndex();
					if (i >= max)
						return;
					
					Phenotype p = pop.getPhenotype(i);
					checkAndCalcFitness(p);
				}
			});
			pool[t].start();
		}
		
		for (int t = 0; t < threads; t++) {
			try {
				pool[t].join();
			} catch (InterruptedException e) {
				throw new RuntimeException("Thread interrupted. Unfinished fitness calculation.");
			}
		}
	}
	
	
	private void staticFitnessCalculation(Population pop) {
		int total = pop.getCount();
		Thread[] pool = new Thread[threads];
		
		int index = 0;
		int part  = total / threads;
		int rem   = total % threads;
		
		for (int t = 0; t < threads; t++) {
			final int start = index;
			
			int tmplen = part;
			if (rem-- > 0) tmplen++;
			index += tmplen;
			
			final int len = tmplen;
			final int id = t;
			
			pool[t] = new Thread(() -> {
				for (int i = start; i < start+len; i++) {
					Phenotype p = pop.getPhenotype(i);
					checkAndCalcFitness(p);
				}
			});
			pool[t].start();
		}
		
		for (int t = 0; t < threads; t++) {
			try {
				pool[t].join();
			} catch (InterruptedException e) {
				throw new RuntimeException("Thread interrupted. Unfinished fitness calculation.");
			}
		}
	}
	
	
	public void abort() {
		aborted = true;
	}
	
	
	public boolean isAborted() {
		return aborted;
	}
	
	
	public void pause() {
		paused = true;
	}
	
	
	public void resume() {
		paused = false;
		synchronized (pauseObj) {
			pauseObj.notifyAll();
		}
	}
	
	
	public boolean isPaused() {
		return paused;
	}
	
	
	public int getCurrentGeneration() {
		return generation;
	}
	
	
	public int getGenerationLimit() {
		return generationLimit;
	}
	
	
	public static synchronized void threadPrintln(String str) {
		System.out.println(str);
	}
	
	
	public EAResults getResults() {
		EAResults res = new EAResults();
		res.fitness_averages   = fitnessAverages;
		res.fitness_maximums   = fitnessBests;
		res.fitness_deviations = fitnessDeviations;
		return res;
	}
	
	/**
	 * Returns a shallow copy of the population. Elements can safely be removed from
	 * the ruturned population, but the elements themselves must not be modyfied before
	 * evolution has ended.
	 * @return 
	 */
	public Population getPopulation() {
		return new Population(population);
	}
	
	
	public void serialize(String filename) throws IOException {
		ObjectOutputStream oos = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(filename)));
		oos.writeObject(this);
		oos.close();
	}
	
	
	public static EA deserialize(String filename) throws IOException, ClassNotFoundException {
		ObjectInputStream ois = new ObjectInputStream(new BufferedInputStream(new FileInputStream(filename)));
		EA ea = (EA) ois.readObject();
		ea.listeners = new ArrayList<>();
		ois.close();
		return ea;
	}
	
	
	protected void printVerbose(String str) {
		if (!quiet && verbose) System.out.println(str);
	}
	
	
	private void checkAndCalcFitness(Phenotype p) {
		boolean evaluated = p.isEvaluated();
		
		if (!evaluated || (evaluated && recalculateFitness)) {
			p.evaluate();
		}
	}
}
