package eval.models;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Iterator;


public class Population implements Iterable<Phenotype>, Serializable {
	
	private final int POP_SIZE;
	private final ArrayList<Phenotype> pop;

	
	public Population(int popSize) {
		this.POP_SIZE = popSize;
		this.pop = new ArrayList<>(popSize);
	}
	
	
	public Population(Population original) {
		this.POP_SIZE = original.POP_SIZE;
		this.pop = new ArrayList<>(original.pop);
	}
	
	
	/**
	 * Returns the phenotype at a given index.
	 * @param index
	 * @return 
	 */
	public Phenotype get(int index) {
		return pop.get(index);
	}
	
	/**
	 * Sort this population according to fitness.
	 * @param ascending 
	 */
	public void sort(boolean ascending) {
		java.util.Collections.sort(pop, ascending ? Comparator.naturalOrder() : Comparator.reverseOrder());
	}
	
	/**
	 * Sort this population according to fitness. 
	 */
	public void sort() {
		java.util.Collections.sort(pop, Comparator.reverseOrder());
	}
	
	
	/**
	 * Returns a sorted, shallow copy of this population.
	 * @param ascending
	 * @return 
	 */
	public Population getSorted(boolean ascending) {
		Population sortedPop = new Population(POP_SIZE);
		sortedPop.pop.addAll(this.pop);
		sortedPop.sort(ascending);
		return sortedPop;
	}
	
	
	/**
	 * Add a phenotype to the population.
	 * @param p 
	 */
	public void add(Phenotype p) {
		if (isFull())
			throw new RuntimeException("Population is full.");
		else
			pop.add(p);
	}
	
	
	public void add(ArrayList<Phenotype> p) {
		if (pop.size() + p.size() > POP_SIZE)
			throw new RuntimeException("Population is full.");
		
		pop.addAll(p);
	}
	
	
	public ArrayList<Phenotype> getPop() {
		return pop;
	}
	
	
	/**
	 * Get the current number of phenotypes in the population.
	 * @return 
	 */
	public int getCount() {
		return pop.size();
	}
	
	
	public int getMaxSize() {
		return POP_SIZE;
	}
	
	
	public boolean isFull() {
		return pop.size() >= POP_SIZE;
	}
	
	
	public Phenotype getPhenotype(int index) {
		return pop.get(index);
	}
	
	
	public ArrayList<Phenotype> getPopClone() {
		return new ArrayList<>(pop);
	}
	
	
	/**
	 * Removes all phenotypes from this population.
	 */
	public void clear() {
		clear(0);
	}
	
	
	/**
	 * Clear this population, but retain elites.
	 * @param elitism 
	 */
	public void clear(int elitism) {
		if (pop.isEmpty())
			return;
		
		if (elitism == 1) {
			Phenotype best = getMaxFitness();
			pop.clear();
			pop.add(best);
		} else if (elitism > 1) {
			sort();
			ArrayList<Phenotype> elitists = new ArrayList<>(elitism);
			
			for (int i = 0; i < elitism; i++)
				elitists.add(pop.get(i));
			
			pop.clear();
			pop.addAll(elitists);
		} else {
			pop.clear();
		}
	}

	
	@Override
	public Iterator<Phenotype> iterator() {
		return pop.iterator();
	}
	
	
	public double getAvgFitness() {
		double sum = 0.0;
		
		for (Phenotype i : pop) {
			sum += i.getFitness();
		}
		
		return sum / (double)pop.size();
	}
	
	
	public double getStdFitness() {
		double s = 0.0;
		double avg = getAvgFitness();
		
		for (Phenotype i : pop) {
			double d = i.getFitness() - avg;
			s += d * d;
		}
		
		return Math.sqrt(s / (double)getCount());
	}
	
	
	public Phenotype getMaxFitness() {
		Phenotype best = pop.get(0);
		
		for (Phenotype p : pop) {
			if (p.getFitness() > best.getFitness())
				best = p;
		}
		
		return best;
	}
	
	
	public Phenotype getMinFitness() {
		Phenotype best = pop.get(0);
		
		for (Phenotype p : pop) {
			if (p.getFitness() < best.getFitness())
				best = p;
		}
		
		return best;
	}

	
	@Override
	public String toString() {
		return "Population{" + "POP_SIZE=" + POP_SIZE + ", count=" + pop.size() + '}';
	}

}
