package eval.util.roulette;

import eval.util.Utils;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;


public class Roulette<T extends RouletteEntry> implements Serializable {

	private final ArrayList<Entry<T>> entries;
	private double p_sum = 0.0;
	private double p_max = Double.MIN_VALUE; 
	private double p_min = Double.MAX_VALUE;
	
	/**
	 * 
	 * @param capacity The starting capacity of the roulette.
	 */
	public Roulette(int capacity) {
		this.entries = new ArrayList<>(capacity);
	}

	
	public Roulette() {
		this(100);
	}
	
	
	public void add(T element, double p) {
		p_max = Math.max(p, p_max);
		p_min = Math.min(p, p_min);
		p_sum += p;
		Entry<T> e = new Entry<>(element, p);
		entries.add(e);
	}
	
	
	public void addAll(List<T> list) {
		addAll(list, true);
	}
	
	
	public void addAll(Iterable<T> iterable) {
		addAll(iterable, true);
	}
	
	
	public void addAll(List<T> list, boolean maximize) {
		if (maximize) {
			list.forEach((e) -> add(e, e.getRouletteValue()));
		}
		else {
			for (RouletteEntry re : list) {
				p_max = Math.max(re.getRouletteValue(), p_max);
				p_min = Math.min(re.getRouletteValue(), p_min);
			}
			list.forEach((e) -> add(e, (p_max + p_min) - e.getRouletteValue()));
		}
	}
	
	
	public void addAll(Iterable<T> iterable, boolean maximize) {
		if (maximize) {
			iterable.forEach((e) -> add(e, e.getRouletteValue()));
		}
		else {
			for (RouletteEntry re : iterable) {
				p_max = Math.max(re.getRouletteValue(), p_max);
				p_min = Math.min(re.getRouletteValue(), p_min);
			}
			iterable.forEach((e) -> add(e, (p_max + p_min) - e.getRouletteValue()));
		}
	}
	
	
	public int numberOfEntries() {
		return entries.size();
	}
	
	
	public T spin(boolean remove) {
		if (entries.isEmpty())
			throw new RuntimeException("No entries in RouletteWheel.");
		
		double r = Utils.randomDouble() * p_sum;
		double cumVal = 0.0;
		
		Entry<T> element = entries.get(0);
		
		for (int i = 0; i < entries.size(); i++) {
			element = entries.get(i);
			cumVal += element.p;
			
			if (r <= cumVal) {
				if (remove) {
					entries.remove(i);
					p_sum -= element.p;
				}
				break;
			}
		}
		
		return element.object;
	}	

	
	@Override
	public String toString() {
		return "Roulette{" + "entries=" + entries + ", p_sum=" + p_sum + ", p_max=" + p_max + ", p_min=" + p_min + '}';
	}

	
	
	
	
	
	
	
	
	static class A implements RouletteEntry {
		int index;
		double val;
		public A(int index, double val) {
			this.index = index;
			this.val = val;
		}
		@Override
		public double getRouletteValue() {
			return val;
		}
	}
	
	
	public static void main(String[] args) {
		A a0 = new A(0, 1);
		A a1 = new A(1, 8);
		A a2 = new A(2, 1);
		
		ArrayList<A> as = new ArrayList<>();
		as.add(a0);
		as.add(a1);
		as.add(a2);
		
		int[] stats = new int[3];
		
		Roulette<A> r = new Roulette<>(3);
		r.addAll(as, false);
		
		r.spin(true);
			
		for (int i = 0; i < 10000; i++) {
			A a = r.spin(false);
			stats[a.index]++;
		}
		
		System.out.println("0: "+stats[0]);
		System.out.println("1: "+stats[1]);
		System.out.println("2: "+stats[2]);
	}
}
