package creativebehaviour.fuzzy;

import creativebehaviour.fuzzy.memberfunc.MembershipFunc;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

/**
 *
 * @author Bjørnar W. Alvestad
 */
public class FuzzySet extends Set {
	
	private int anonValueCounter = 0;
	
	private double min;
	private double max;
	
	/** Linguistic values (fuzzy set). */
	private final List<FuzzyValue> values;

	
	public FuzzySet(int var, double min, double max) {
		super(var);
		this.min = min;
		this.max = max;
		this.values = new ArrayList<>();
	}
	
	
	public FuzzySet(String varName, double min, double max) {
		this(Variable.indexOf(varName), min, max);
	}
	
	
	/**
	 * Copy constructor.
	 * @param original 
	 */
	public FuzzySet(FuzzySet original) {
		super(original.var);
		this.min = original.min;
		this.max = original.max;
		this.values = new ArrayList<>(original.values.size());
		
		for (FuzzyValue v : original.values) {
			this.values.add(new FuzzyValue(v));
		}
	}
	
	
	public void addFuzzyValue(String name, MembershipFunc memFunc) {
		values.add(new FuzzyValue(name, memFunc));
	}
	
	
	public void addFuzzyValue(MembershipFunc memFunc) {
		addFuzzyValue(getName() + "-" + anonValueCounter++, memFunc);
	}
	
	
	public List<FuzzyValue> getValues() {
		return values;
	}
	
	
	public int numVals() {
		return values.size();
	}
	
	
	/**
	 * Creates a double[] with the appropriate length to be used as a buffer for this variable.
	 * @return 
	 */
	public double[] createBuffer() {
		return new double[numVals()];
	}
	
	
	/**
	 * Reads this fuzzy set at the specified x-position.<br>
	 * Results are placed in the supplied buffer.
	 * 
	 * @param x
	 * @param outBuff 
	 */
	public void read(double x, double[] outBuff) {
		if (outBuff.length < values.size())
			throw new IllegalArgumentException("Buffer provided is too small. Need " + values.size() + ", but was " + outBuff.length);
		
		for (int i = 0; i < values.size(); i++) {
			outBuff[i] = values.get(i).membershipFunction.f(x);
		}
	}

	
	public void normalize() {
		double normalFactor = Math.max(Math.abs(min), Math.abs(max));
		
		min /= normalFactor;
		max /= normalFactor;
		
		for (FuzzyValue v : values) {
			v.membershipFunction.normalize(normalFactor);
		}
	}
	
	
	@Override
	public String toString() {
		return "fuzzyset:" + getName();
//		StringBuilder str = new StringBuilder("Variable{" + "min=" + min + ", max=" + max + ", name=" + getName() + "}\n");
//		for (FuzzyValue v : values) {
//			str.append(v.toString()).append("\n");
//		}
//		return str.toString();
	}
	
	
	public void createDebugData() {
		File folder = new File("debug");
		if (!folder.exists()) folder.mkdir();
		
		try {
			PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter("debug/" + getName() + ".txt", false)));
			
			for (int i = 0; i < values.size(); i++) {
				FuzzyValue fv = values.get(i);
				pw.print("\t");
				pw.print(fv.linguisticName);
			}
			pw.println();
			
			double[] buff = createBuffer();
			final int samplePoints = 100;
			final double dx = (max - min) / (double)samplePoints;
			
			for (double x = min; x <= max; x += dx) {
				if (x == min) pw.print(x);
				pw.print("\t");
				read(x, buff);
				for (int j = 0; j < buff.length; j++) {
					pw.printf(Locale.ENGLISH, "%.2f", buff[j]);
					if (j != buff.length-1) pw.print("\t");
				}
				pw.println();
			}
			
			pw.print(max);
			pw.print("\t");
			read(max, buff);
			for (int j = 0; j < buff.length; j++) {
				pw.printf(Locale.ENGLISH, "%.2f", buff[j]);
				if (j != buff.length-1) pw.print("\t");
			}
			pw.println();
			
			

			pw.flush();
			pw.close();
		} catch (Exception e) {
			System.err.println(e);
		}
		
	}
	
	
}
