package creativebehaviour.fuzzy;

import creativebehaviour.ents.misc.EntType;
import static creativebehaviour.ents.misc.EntType.*;
import static creativebehaviour.fuzzy.Variable.Type.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

/**
 * Immutable class.
 * @author Bjørnar W. Alvestad
 */
public class Variable {
	
	private static final List<Variable> variables = new ArrayList<>();
	private static final Map<EntType, List<Variable>> entTypeVarMap = new HashMap<>();
	private static final Map<String, Variable> nameVarMap = new HashMap<>();
	private static final Map<Variable, Integer> varIndexMap = new HashMap<>();
	
	/**
	 * Dummy method to force initialization of the class.
	 */
	public static void load() {
	}
	
	static {
		for (EntType t : EntType.values()) {
			entTypeVarMap.put(t, new ArrayList<>());
		}
		
		// Actions
		/*0*/ addVar("Move",              BODY,         ACTION, "Reverse", "None",     "Forward");
		/*1*/ addVar("Turn",              BODY,         ACTION, "CCV",     "None",     "CV");
		/*2*/ addVar("Shoot",             MOUTH,        ACTION, "No",      "Yes");    
		/*3*/ addVar("Rotate",            ARM,          ACTION, "CCV",     "None",     "CV");
		/*4*/ addVar("Push",              MOTOR,        ACTION, "No",      "Yes");
		 
		// Sensors 
		/*5*/ addVar("HP",                BODY,         SENSOR, "Low",     "Moderate", "High");
		/*6*/ addVar("Distance", "Dist",  EYE,          SENSOR, "Close",   "Medium",   "Far", "OutOfRange");
		/*7*/ addVar("Direction", "Dir",  EYE,          SENSOR, "Left",    "Center",   "Right");
		/*8*/ addVar("Noise",             EAR,          SENSOR, "None",    "Quiet",    "Loud");
		/*9*/ addVar("Smell",             NOSE,         SENSOR, "Bad",     "None",     "Good");
		/*10*/addVar("Touch",             TOUCH_SENSOR, SENSOR, "None",    "Light",    "Hard");
		/*11*/addVar("FoodEaten", "Food", MOUTH,        SENSOR, "No",      "Yes");
		
		EntType.createIOLegends();
	}
	
	
	private static void addVar(String name, EntType entType, Type varType, String... values) {
		addVar(name, name.substring(0, Math.min(6, name.length())), entType, varType, values);
	}
	
	
	private static void addVar(String name, String shortName, EntType entType, Type varType, String... values) {
		Variable var = new Variable(name, shortName, entType, varType, values);
		varIndexMap.put(var, variables.size());
		variables.add(var);
		entTypeVarMap.get(entType).add(var);
		
		if (nameVarMap.containsKey(name)) {
			System.err.println("Variable \"" + name + "\" already exists!");
			System.exit(1);
		} else {
			nameVarMap.put(name, var);
		}
	}
	
	
	public static Variable getVar(int index) {
		return variables.get(index);
	}
	
	
	public static List<Variable> getVarsForEntity(EntType t) {
		return new ArrayList<>(entTypeVarMap.get(t));
	}
	
	
	public static Variable getVar(String name) {
		if (!nameVarMap.containsKey(name)) {
			System.err.println("No variable named \"" + name + "\".");
			System.exit(1);
		}
		return nameVarMap.get(name);
	}
	
	
	public static Variable[] getVars(String... names) {
		Variable[] vars = new Variable[names.length];
		for (int i = 0; i < names.length; i++) {
			vars[i] = getVar(names[i]);
		}
		return vars;
	}
	
	
	public static int indexOf(Variable var) {
		return varIndexMap.get(var);
	}
	
	
	public static int indexOf(String varName) {
		return indexOf(getVar(varName));
	}
	
	
	public enum Type {
		SENSOR, ACTION;
	}
	
	
	
	public final String name;
	public final String shortName;
	public final EntType entType;
	public final Type varType;
	public final int id;
	private final Value[] values;
	
	
	private Variable(String name, String shortName, EntType entType, Type varType, String... values) {
		this.name = name;
		this.shortName = shortName;
		this.entType = entType;
		this.values = new Value[values.length];
		this.varType = varType;
		
		int hash = 3;
		hash = 89 * hash + Objects.hashCode(this.name);
		hash = 89 * hash + Objects.hashCode(this.entType);
		this.id = hash;
		
		for (int i = 0; i < values.length; i++) {
			this.values[i] = new Value(values[i]);
		}
	}

	
	@Override
	public String toString() {
		return "var:" + name;
		//return "Variable{" + "name=" + name + ", entType=" + entType + ", varType=" + varType + ", values=" + Arrays.toString(values) + '}';
	}
	
	
	public Value[] getValuesCopy() {
		Value[] cpy = new Value[values.length];
		System.arraycopy(values, 0, cpy, 0, values.length);
		return cpy;
	}
	
	
	/**
	 * Do not modify returned array.
	 * @return 
	 */
//	public Value[] getValues() {
//		return values;
//	}
	
	
	public Value getValue(int i) {
		return values[i];
	}

	
	public int getNumValues() {
		return values.length;
	}

	
	@Override
	public int hashCode() {
		return id;
	}

	
	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final Variable other = (Variable) obj;
		return this.id == other.id;
	}
	
	
	
	public class Value {
		public final String name;
		
		public Value(String name) {
			this.name = name;
		}

		@Override
		public String toString() {
			return name;
		}
		
		public Variable getParentVariable() {
			return Variable.this;
		}
	}

}
