package creativebehaviour.fuzzy.memberfunc;

import java.util.Arrays;

/**
 *
 * @author Bjørnar W. Alvestad
 */
public class Grade extends MembershipFunc implements Reversable {
	
	/** Grade slope. */
	private double a;
	
	/** Grade center. */
	private double c;
	
	/** Flag to indicate reversed grade. */
	private final boolean reverse;
	
	/** Grade x-points. */
	private final double[] xp = new double[2];

	
	/**
	 * Grade:
	 * 
	 *           _____
	 *        | /x1
	 *         / 
	 *      a |
	 *       /
	 * _____/.|......
	 *     x0 c         
	 * 
	 * @param a Grade slope.
	 * @param c Grade center.
	 * @param reverse Flag to indicate reversed grade.
	 */
	Grade(double a, double c, boolean reverse) {
		super(MembershipFunctions.GRADE);
		if (a <= 0) throw new IllegalArgumentException("Slope less than or equal to 0.");
		
		this.a = a;
		this.c = c;
		this.reverse = reverse;
		
		xp[0] = c - 0.5/a;
		xp[1] = c + 0.5/a;
	}
	
	
	/**
	 * 
	 * @param x
	 * @param reverse
	 */
	Grade(double[] x, boolean reverse) {
		super(MembershipFunctions.GRADE);
		if (x.length != 2) throw new IllegalArgumentException("Length of x-points table passed to Grade must be 2. Is " + x.length + ".");
		for (int i = 1; i < x.length; i++) {
			if (x[i] < x[i-1]) throw new IllegalArgumentException("x-points table passed to Grade must be strictly increasing. Is " + Arrays.toString(x) + ".");
		}
		
		System.arraycopy(x, 0, this.xp, 0, 2);
		this.c = (x[0] + x[1])*0.5;
		this.reverse = reverse;
		
		if (Double.compare(x[0], x[1]) == 0) a = reverse ? Double.NEGATIVE_INFINITY : Double.POSITIVE_INFINITY;
		else a = 1.0 / (x[1] - x[0]);
	}

	
	Grade(Grade original) {
		super(MembershipFunctions.GRADE);
		
		this.a = original.a;
		this.c = original.c;
		this.reverse = original.reverse;
		
		this.xp[0] = original.xp[0];
		this.xp[1] = original.xp[1];
	}

	
	@Override
	public double f(double x) {
		if (x <= xp[0]) x = 0;
		else if (x < xp[1]) x = a*(x-xp[1]) + 1;
		else x = 1;
		return reverse ? 1 - x : x;
	}
	

	@Override
	public String toString() {
		return super.toString() + "{" + "a=" + a + ", c=" + c + ", reverse=" + reverse + ", x0=" + xp[0] + ", x1=" + xp[1] + '}';
	}

	
	@Override
	public void normalize(double normalizeFactor) {
		c /= normalizeFactor;
		for (int i = 0; i < xp.length; i++) {
			xp[i] /= normalizeFactor;
		}
		if (Double.isFinite(a)) a = 1.0 / (xp[1] - xp[0]);
	}

	
	@Override
	public MembershipFunc copy() {
		return new Grade(this);
	}

	
	@Override
	public MembershipFunc getReversed() {
		return new Grade(xp, !reverse);
	}
	
}
