package creativebehaviour.fuzzy.memberfunc;

import java.util.Arrays;

/**
 *
 * @author Bjørnar W. Alvestad
 */
public class Trapezoid extends MembershipFunc {
	
	/** The left slope of the trapezoid. */
	private double a1;
	
	/** The right slope of the trapezoid. */
	private double a2;
	
	/** The center of the trapezoid. f(c) == 1 */
	private double c;
	
	/** The size if the flat top of the trapezoid. */
	private double top;
	
	/** The 'radius' of the base of the trapezoid. */
	private double r;
	
	/** Trapezoid x-points. */
	private final double[] xp = new double[4];
	
	
	/**
	 * Trapezoid:
	 * 
	 *         top
	 *       x1 __|__ x2
	 *         /     \
	 *     a1 /   |   \ a2
	 *       /         \
	 * _____/.....|.....\_____
	 *     x0     c     x3
	 * 
	 * @param a Trapezoid slope.
	 * @param c Trapezoid center.
	 * @param top Size of top section of the trapezoid.
	 */
	Trapezoid(double a, double c, double top) {
		this(a, c, top, MembershipFunctions.TRAPEZOIDAL);
	}
	
	
	/**
	 * 
	 * @param a
	 * @param c
	 * @param top
	 * @param memFunEnum 
	 */
	Trapezoid(double a, double c, double top, MembershipFunctions memFunEnum) {
		super(memFunEnum);
		if (a <= 0) throw new IllegalArgumentException("Slope less than or equal to 0.");
		
		this.a1 = a;
		this.a2 = a;
		this.c = c;
		this.top = top;
		
		r = 1/a + top;
		
		xp[0] = c - r;
		xp[1] = c - top;
		xp[2] = c + top;
		xp[3] = c + r;
	}

	
	/**
	 * 
	 * @param x 
	 */
	Trapezoid(double[] x) {
		this(x, MembershipFunctions.TRAPEZOIDAL);
	}
	
	
	/**
	 * 
	 * @param x 
	 * @param memFunEnum 
	 */
	Trapezoid(double[] x, MembershipFunctions memFunEnum) {
		super(memFunEnum);
		if (x.length != 4) throw new IllegalArgumentException("Length of x-points table passed to Trapezoid must be 4. Is " + x.length + ".");
		for (int i = 1; i < x.length; i++) {
			if (x[i] < x[i-1]) throw new IllegalArgumentException("x-points table passed to Trapezoid must be strictly increasing. Is " + Arrays.toString(x) + ".");
		}
		
		System.arraycopy(x, 0, xp, 0, 4);
		this.c = (x[0] + x[3])*0.5;
		this.r = (x[3] - x[0])*0.5;
		this.top = (x[2] - x[1])*0.5;
		
		if (Double.compare(x[0], x[1]) == 0) a1 = Double.POSITIVE_INFINITY;
		else a1 = 1.0 / (x[1] - x[0]);
		
		if (Double.compare(x[2], x[3]) == 0) a2 = Double.POSITIVE_INFINITY;
		else a2 = -1.0 / (x[3] - x[2]);
	}
	
	
	Trapezoid(Trapezoid original) {
		this(original.xp, original.memFunEnum);
	}
	
	
	@Override
	public double f(double x) {
		if (x <= xp[0]) return 0;
		if (x < xp[1]) return a1*(x-xp[1]) + 1;
		if (x <= xp[2]) return 1;
		if (x < xp[3]) return a2*(x-xp[2]) + 1;
		return 0;
	}

	
	@Override
	public String toString() {
		return super.toString() + "{" + "a1=" + a1 + ", a2=" + a2 + ", c=" + c + ", top=" + top + ", r=" + r + ", x0=" + xp[0] + ", x1=" + xp[1] + ", x2=" + xp[2] + ", x3=" + xp[3] + '}';
	}

	
	@Override
	public void normalize(double normalizeFactor) {
		for (int i = 0; i < xp.length; i++) {
			xp[i] /= normalizeFactor;
		}
		
		this.c = (xp[0] + xp[3])*0.5;
		this.r = (xp[3] - xp[0])*0.5;
		this.top = (xp[2] - xp[1])*0.5;
		
		if (Double.compare(xp[0], xp[1]) == 0) a1 = Double.POSITIVE_INFINITY;
		else a1 = 1.0 / (xp[1] - xp[0]);
		
		if (Double.compare(xp[2], xp[3]) == 0) a2 = Double.POSITIVE_INFINITY;
		else a2 = -1.0 / (xp[3] - xp[2]);
	}

	
	@Override
	public MembershipFunc copy() {
		return new Trapezoid(this);
	}
	
}
