package creativebehaviour.fuzzy.rule;

import java.util.ArrayList;

/**
 *
 * @author Bjørnar W. Alvestad
 */
public class ChainRule extends Rule {
	
	public Operation operation;
	
	
	public ChainRule(ArrayList<RuleClause> clauses, RuleClause thenClause, Operation operation) {
		super(clauses, thenClause);
		this.operation = operation;
	}
	
	
	public ChainRule(ChainRule original) {
		super(original);
		this.operation = original.operation;
	}

	
	@Override
	public String toString() {
		StringBuilder str = new StringBuilder("(");
		str.append(clauses.size()).append(") IF ");
		for (int i = 0; i < clauses.size(); i++) {
			str.append("(").append(clauses.get(i).toString()).append(")");
			if (i < clauses.size() - 1) str.append(" ").append(operation.name()).append(" ");
		}
		str.append(" THEN (").append(thenClause.toString()).append(")");
		return str.toString();
	}

	
	@Override
	public Rule copy() {
		return new ChainRule(this);
	}
	
	
	@Override
	public double eval(double[][] fuzzyData) {
		double res = 0.0;
		
		for (RuleClause clause : clauses) {
			try {
				double val = fuzzyData[clause.entitySensorIndex][clause.value];
				res = operation.evaluate(res, val);
				if (clause.negation) res = 1 - res;
			} catch (Exception e) {
				System.err.println("Index was " + clause.entitySensorIndex + "," + clause.value + ". Array Length is " + fuzzyData.length + "," + fuzzyData[clause.entitySensorIndex].length);
				throw e;
			}
		}
		
		return res;
	}
	
	
	
	public static class Builder {
		private final ArrayList<RuleClause> clauses;
		private RuleClause thenClause;
		private Operation operation;

		/**
		 * Creates a new builder.
		 */
		public Builder() {
			this.clauses = new ArrayList<>();
		}
		
		/**
		 * Creates a new builder to extend an existing ChainRule.
		 * @param original 
		 */
		public Builder(ChainRule original) {
			this();
			clauses.addAll(original.clauses);
			thenClause = original.thenClause;
		}
		
		public Builder addClause(RuleClause clause) {
			clauses.add(clause);
			return this;
		}
		
		public Builder setThenClause(RuleClause clause) {
			thenClause = clause;
			return this;
		}
		
		public Builder setOpereation(Operation operation) {
			this.operation = operation;
			return this;
		}
		
		public ChainRule createRule() {
			if (clauses.isEmpty() || thenClause == null || operation == null) {
				throw new RuntimeException("All fields not set.");
			}
			return new ChainRule(clauses, thenClause, operation);
		}
	}
	
	
}
