package creativebehaviour.fuzzy.sample;

import java.util.ArrayList;

class Reasoner {

	public static void main(String[] args) {
		double distance = 3.7;
		double delta = 1.1;
		
		Variable distVar = new Variable(0, 10, 0.05);
		distVar.addFuzzySet(FuzzySet.Type.ReverseGrade, 1.0, 2.5);
		distVar.addFuzzySet(FuzzySet.Type.Triangle, 1.5, 3.0, 4.5);
		distVar.addFuzzySet(FuzzySet.Type.Triangle, 3.5, 5.0, 6.5);
		distVar.addFuzzySet(FuzzySet.Type.Triangle, 5.5, 7.0, 8.5);
		distVar.addFuzzySet(FuzzySet.Type.Grade, 7.5, 9.0);
		
		Variable deltaVar = new Variable(-5, 5, 0.05);
		deltaVar.addFuzzySet(FuzzySet.Type.ReverseGrade, -4.0, -2.5);
		deltaVar.addFuzzySet(FuzzySet.Type.Triangle, -3.5, -2.0, -0.5);
		deltaVar.addFuzzySet(FuzzySet.Type.Triangle, -1.5, 0.0, 1.5);
		deltaVar.addFuzzySet(FuzzySet.Type.Triangle, 0.5, 2.0, 3.5);
		deltaVar.addFuzzySet(FuzzySet.Type.Grade, 2.5, 4.0);
		
		Variable actionVar = new Variable(-10, 10, 0.01);
		actionVar.addFuzzySet(FuzzySet.Type.ReverseGrade, -8.0, -5.0);
		actionVar.addFuzzySet(FuzzySet.Type.Triangle, -7.0, -4.0, -1.0);
		actionVar.addFuzzySet(FuzzySet.Type.Triangle, -3.0, 0.0, 3.0);
		actionVar.addFuzzySet(FuzzySet.Type.Triangle, 1.0, 4.0, 7.0);
		actionVar.addFuzzySet(FuzzySet.Type.Grade, 5.0, 8.0);
		
		double[] distfuzz = distVar.fuzzification(distance);
		double[] deltafuzz = deltaVar.fuzzification(delta);
		double[] actionfuzz = new double[5];
		
		actionfuzz[0] = Rules.ruleC1(distfuzz, deltafuzz);
		actionfuzz[1] = Rules.ruleC2(distfuzz, deltafuzz);
		actionfuzz[2] = Rules.ruleC3(distfuzz, deltafuzz);
		actionfuzz[3] = Rules.ruleC4(distfuzz, deltafuzz);
		actionfuzz[4] = Rules.ruleC5(distfuzz, deltafuzz);
		
		double action = actionVar.defuzzification(actionfuzz);
		System.out.println(action);
	}
}

class Rules {
// distance delta

	public static double ruleC3(double[] x, double[] y) {
		return Math.min(x[1], y[3]);
	}

	public static double ruleC2(double[] x, double[] y) {
		return Math.min(x[1], y[2]);
	}

	public static double ruleC4(double[] x, double[] y) {
		return Math.min(x[2], y[3]);
	}

	public static double ruleC5(double[] x, double[] y) {
		return Math.min(x[4], Math.max(y[3], y[4]));
	}

	public static double ruleC1(double[] x, double[] y) {
		return x[0];
	}
}

class Variable {

	public final ArrayList<FuzzySet> sets = new ArrayList<>();
	public double[] values;
	public double max;
	public double min;
	public double step;

	public Variable(double min, double max, double step) {
		this.max = max;
		this.min = min;
		this.step = step;
	}

	public void addFuzzySet(FuzzySet.Type type, double x0, double x1, double x2) {
		sets.add(new FuzzySet(type, x0, x1, x2));
	}

	public void addFuzzySet(FuzzySet.Type type, double x0, double x1) {
		if (type == FuzzySet.Type.Triangle) {
			throw new RuntimeException("Cant pass TRIANGLE to this method.");
		}
		addFuzzySet(type, x0, x1, 0.0);
	}

	public double[] fuzzification(double in) {
		double[] res = new double[5];
		int c = 0;
		for (FuzzySet fs : sets) {
			res[c++] = fs.eval(in, 1.0);
		}
		values = res;
		return res;
	}

	public double defuzzification(double[] fuzzySetVals) {
		double val_sum = 0.0;
		double num = 1.0;
		for (double sample_x = min; sample_x < max; sample_x += step) {
			double value = 0.0;
			for (int i = 0; i < sets.size(); i++) {
				double fs_val = sets.get(i).eval(sample_x, fuzzySetVals[i]);
				value = Math.max(value, fs_val);
				val_sum += fs_val;
			}
			num += value * sample_x;
		}
		return num / val_sum;
	}
}

class FuzzySet {

	public enum Type {
		Triangle, Grade, ReverseGrade
	}
	
	public final Type type;
	public final double x0, x1, x2;

	public FuzzySet(Type type, double x0, double x1, double x2) {
		this.type = type;
		this.x0 = x0;
		this.x1 = x1;
		this.x2 = x2;
	}

	public double eval(double x, double clip) {
		if (type == Type.Triangle) {
			return triangle(x, x0, x1, x2, clip);
		}
		if (type == Type.Grade) {
			return grade(x, x0, x1, clip);
		}
		if (type == Type.ReverseGrade) {
			return reverse_grade(x, x0, x1, clip);
		}
		throw new RuntimeException("Something went wrong.");
	}

	private static double triangle(double position, double x0, double x1, double x2, double clip) {
		double value = 0.0;
		if (position >= x0 && position <= x1) {
			value = (position - x0) / (x1 - x0);
		} else if (position >= x1 && position <= x2) {
			value = (x2 - position) / (x1 - x0);
		}
		if (value > clip) {
			value = clip;
		}
		return value;
	}

	private static double grade(double position, double x0, double x1, double clip) {
		double value = 0.0;
		if (position >= x1) {
			value = 1.0;
		} else if (position <= x0) {
			value = 0.0;
		} else {
			value = (position - x0) / (x1 - x0);
		}
		if (value > clip) {
			value = clip;
		}
		return value;
	}

	private static double reverse_grade(double position, double x0, double x1, double clip) {
		double value = 0.0;
		if (position <= x0) {
			value = 1.0;
		} else if (position >= x1) {
			value = 0.0;
		} else {
			value = (x1 - position) / (x1 - x0);
		}
		if (value > clip) {
			value = clip;
		}
		return value;
	}
}
