package creativebehaviour.models.fuzzy;

import creativebehaviour.fuzzy.FuzzySet;
import creativebehaviour.fuzzy.SugenoSet;
import creativebehaviour.fuzzy.rule.Rule;
import creativebehaviour.fuzzy.rule.RuleClause;
import creativebehaviour.models.Creature;
import java.util.Arrays;

/**
 *
 * @author Bjørnar W. Alvestad
 */
public class FuzzyCreature extends Creature {
	
	/** The fuzzy genotype. */
	private final FuzzyGenotype fuzzyGenotype;
	
	/** Fuzzy sets that match up with input sensor data in tick. */
	private final FuzzySet[] sensorSets;
	
	/** Sugeno sets that match up with output actuator data in tick. */
	private final SugenoSet[] actuatorSets;
	
	/** Buffer for each sensor input. */
	private final double[][] fuzzyData;
	
	/** Buffer for each actuator output value. */
	private final double[][] outputData;
	
	private final int numSensors;
	private final int numActuators;
	
	
	public FuzzyCreature(FuzzyGenotype fuzzyGenotype) {
		super(fuzzyGenotype.treeGenotype, fuzzyGenotype);
		this.fuzzyGenotype = fuzzyGenotype;
		
		this.numSensors = sensorLegend.length;
		this.numActuators = actuatorLegend.length;
		
		this.sensorSets = new FuzzySet[numSensors];
		this.actuatorSets = new SugenoSet[numActuators];
		
		this.fuzzyData = new double[numSensors][];
		this.outputData = new double[numActuators][];
		
		for (int i = 0; i < numSensors; i++) {
			sensorSets[i] = FuzzyGenotype.standardFuzzySets.get(sensorLegend[i].getVar());
			fuzzyData[i] = sensorSets[i].createBuffer();
		}
		
		for (int i = 0; i < numActuators; i++) {
			actuatorSets[i] = FuzzyGenotype.standardSugenoSets.get(actuatorLegend[i].getVar());
			outputData[i] = new double[actuatorLegend[i].getVar().getNumValues()];
		}
		
	}
	
	
	/**
	 * Copy constructor.
	 * @param original 
	 */
	public FuzzyCreature(FuzzyCreature original) {
		this(new FuzzyGenotype(original.fuzzyGenotype));
		setEqual(original);
	}
	
	
	@Override
	public Creature copy() {
		return new FuzzyCreature(this);
	}
	
	
	/**
	 *
	 * @param sensorDataIn
	 * @param actuatorDataOut
	 */
	@Override
	protected void customCreatureTick(float[] sensorDataIn, float[] actuatorDataOut) {
		assert sensorDataIn.length == numSensors;
		assert actuatorDataOut.length == numActuators;
		
		// Clear buffers
		for (int i = 0; i < outputData.length; i++) {
			Arrays.fill(outputData[i], 0.0);
		}
		
		// Get fuzzified inputs from the fuzzy sets.
		for (int i = 0; i < numSensors; i++) {
			sensorSets[i].read(sensorDataIn[i], fuzzyData[i]);
		}
		
		// Evaluate rules.
		for (int i = 0; i < fuzzyGenotype.rules.size(); i++) {
			Rule r = fuzzyGenotype.rules.get(i);
			double ruleRes = r.eval(fuzzyData);
			RuleClause thenClause = r.getThenClause();
			outputData[thenClause.entitySensorIndex][thenClause.value] += ruleRes;
		}
		
		// Defuzzification
		for (int i = 0; i < numActuators; i++) {
			double[] weights = outputData[i];
			double defuzz = actuatorSets[i].eval(weights);
			actuatorDataOut[i] = (float)defuzz;
		}
		
		//System.out.println(Utils.toString(sensorDataIn) + " -> " + Utils.toString(actuatorDataOut));
	}
	
	
	@Override
	public FuzzyGenotype getGenotype() {
		return fuzzyGenotype;
	}

	
	@Override
	public String toString() {
		return getClass().getSimpleName() + "@" + hashCode() + "; rules=" + fuzzyGenotype.rules.size();
	}


	
	
	
}
