package creativebehaviour.models.fuzzy;

import creativebehaviour.ents.wrapper.Wrapper;
import creativebehaviour.fuzzy.FuzzySet;
import creativebehaviour.fuzzy.SetContainer;
import creativebehaviour.fuzzy.SugenoSet;
import creativebehaviour.fuzzy.Variable;
import static creativebehaviour.fuzzy.memberfunc.MembershipFunctions.GRADE;
import static creativebehaviour.fuzzy.memberfunc.MembershipFunctions.TRAPEZOIDAL;
import creativebehaviour.fuzzy.rule.ChainRule;
import creativebehaviour.fuzzy.rule.Operation;
import creativebehaviour.fuzzy.rule.Rule;
import creativebehaviour.fuzzy.rule.RuleClause;
import creativebehaviour.models.BehaviourGenotype;
import creativebehaviour.models.physical.TreeGenotype;
import creativebehaviour.util.Utils;
import eval.models.Crossover;
import eval.models.Genotype;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

/**
 *
 * @author Bjørnar W. Alvestad
 */
public class FuzzyGenotype extends BehaviourGenotype {
	
	public final TreeGenotype treeGenotype;
	public final ArrayList<Rule> rules;
	public final SetContainer<FuzzySet> fuzzySets;
	public final SetContainer<SugenoSet> sugenoSets;
	
	private final int[] sensorVars;
	private final int[] actuatorVars;
	
	
	public FuzzyGenotype(TreeGenotype treeg, ArrayList<Rule> rules) {
		this.treeGenotype = treeg;
		this.rules = rules == null ? new ArrayList<>() : new ArrayList<>(rules);
		this.fuzzySets = new SetContainer<>();
		this.sugenoSets = new SetContainer<>();
		
		ArrayList<Integer> sv = new ArrayList<>();
		ArrayList<Integer> av = new ArrayList<>();
		
		parseSensorActuatorVars(sv, av);
		
		this.sensorVars = new int[sv.size()];
		this.actuatorVars = new int[av.size()];
		
		for (int i = 0; i < sensorVars.length; i++) {
			sensorVars[i] = sv.get(i);
		}
		for (int i = 0; i < actuatorVars.length; i++) {
			actuatorVars[i] = av.get(i);
		}
	}
	
	
	public FuzzyGenotype(TreeGenotype treeg) {
		this(treeg, null);
		
		// Generate rules
		for (int actionVarIndex = 0; actionVarIndex < actuatorVars.length; actionVarIndex++) {
			Variable actionVar = Variable.getVar(actuatorVars[actionVarIndex]);
			
			for (int actionValue = 0; actionValue < actionVar.getNumValues(); actionValue++) {
				int numRules = Utils.randomInt(1, 4); // TODO: hardcoded number of fuzzy rules
				for (int j = 0; j < numRules; j++) {
					rules.add(generateRandomRule(actionVarIndex, actionValue));
				}
			}
		}
		
	}
	
	
	/**
	 * Copy constructor.
	 * @param original 
	 */
	public FuzzyGenotype(FuzzyGenotype original) {
		this(new TreeGenotype(original.treeGenotype), null);
		
		original.rules.forEach(r -> this.rules.add(r.copy()));
		original.fuzzySets.getSets().forEach(f -> this.fuzzySets.addSet(new FuzzySet(f)));
		original.sugenoSets.getSets().forEach(s -> this.sugenoSets.addSet(new SugenoSet(s)));
	}
	
	
	private void parseSensorActuatorVars(ArrayList<Integer> sv, ArrayList<Integer> av) {
		ArrayList<Wrapper> sensorWraps = new ArrayList<>();
		ArrayList<Wrapper> actuatorWraps = new ArrayList<>();
		
		for (Wrapper w : treeGenotype.tree) {
			if (w.isSensor()) sensorWraps.add(w);
			if (w.isActuator()) actuatorWraps.add(w);
		}
		
		for (Wrapper inw : sensorWraps) {
			Variable.getVarsForEntity(inw.entType).forEach(e -> {
				if (e.varType == Variable.Type.SENSOR) {
					sv.add(Variable.indexOf(e));
				}
			});
		}
		
		for (Wrapper outw : actuatorWraps) {
			Variable.getVarsForEntity(outw.entType).forEach(e -> {
				if (e.varType == Variable.Type.ACTION) {
					av.add(Variable.indexOf(e));
				}
			});
		}
	}
	
	
	private Rule generateRandomRule() {
		return generateRandomRule(Utils.randomInt(actuatorVars.length));
	}
	
	
	private Rule generateRandomRule(int actionVarIndex) {
		return generateRandomRule(actionVarIndex, Utils.randomInt(Variable.getVar(actuatorVars[actionVarIndex]).getNumValues()));
	}
	
	
	private Rule generateRandomRule(int actionVarIndex, int actionVarValue) {
		ChainRule.Builder ruleBuilder = new ChainRule.Builder();
		Variable actionVar = Variable.getVar(actuatorVars[actionVarIndex]);

		int numClauses = Math.round(Math.abs((float)Utils.randomGausian(0, 1)))+1;
		for (int i = 0; i < numClauses; i++) {
			int inVarIndex = Utils.randomInt(sensorVars.length);
			Variable inVar = Variable.getVar(sensorVars[inVarIndex]);
			ruleBuilder.addClause(new RuleClause(Variable.indexOf(inVar), inVarIndex, Utils.randomInt(inVar.getNumValues()), Utils.randomBoolean()));
		}

		ruleBuilder.setOpereation(Utils.randomBoolean() ? Operation.AND : Operation.OR);
		ruleBuilder.setThenClause(new RuleClause(Variable.indexOf(actionVar), actionVarIndex, actionVarValue)); // TODO: add multiple then-clauses to rule

		return ruleBuilder.createRule();
	}
	
	
	@Override
	public Genotype copy() {
		return new FuzzyGenotype(this);
	}

	
	@Override
	public Genotype crossover(Genotype g, Crossover type) {
		FuzzyGenotype other = (FuzzyGenotype)g;
		
		if (!treeGenotype.wrapperEqual(other.treeGenotype)) {
			throw new RuntimeException("Tree genotypes not wrapper equal.");
		}
		
		switch (type) {
			case ONE_POINT:
				return onePointCrossover(other);
			case UNIFORM:
				return uniformCrossover(other);
			default:
				throw new AssertionError(type.name());
		}
	}

	
	private FuzzyGenotype onePointCrossover(FuzzyGenotype other) {
		int aNumRules = this.rules.size();
		int bNumRules = other.rules.size();
		
		int splitIndex = Utils.randomInt(Math.max(aNumRules, bNumRules) + 1);
		
		ArrayList<Rule> newRules = new ArrayList<>(Math.max(aNumRules, bNumRules));
		
		// Add from this
		for (int i = 0; i < splitIndex && i < this.rules.size(); i++) {
			newRules.add(this.rules.get(i));
		}
		
		// Add from other
		for (int i = splitIndex; i < other.rules.size(); i++) {
			newRules.add(other.rules.get(i));
		}
		
		return new FuzzyGenotype(treeGenotype, newRules);
	}
	
	
	private FuzzyGenotype uniformCrossover(FuzzyGenotype other) {
		int maxRules = Math.max(this.rules.size(), other.rules.size());
		ArrayList<Rule> newRules = new ArrayList<>(maxRules);
		
		for (int i = 0; i < maxRules; i++) {
			// Add from this
			if (Utils.randomBoolean()) {
				if (i < this.rules.size()) newRules.add(this.rules.get(i));
			}
			// Add from other
			else {
				if (i < other.rules.size()) newRules.add(other.rules.get(i));
			}
		}
		
		return new FuzzyGenotype(treeGenotype, newRules);
	}
	
	
	@Override
	public void mutate(double prob) {
		// Mutate rules
		for (Rule rule : rules) {
			for (RuleClause clause : rule.getClauses()) {
				mutateRuleClause(clause, prob);
			}
			mutateRuleClause(rule.getThenClause(), prob);
		}
		
		// Add/remove rules
		if (Utils.randomEvent(prob)) {
			// Add rule
			if (Utils.randomBoolean()) {
				rules.add(generateRandomRule());
			}
			// Remove rule
			else if (!rules.isEmpty()) {
				rules.remove(Utils.randomInt(rules.size()));
			}
		}
	}
	
	
	/**
	 * Helper method.
	 * @param clause
	 * @param prob 
	 */
	private static void mutateRuleClause(RuleClause clause, double prob) {
		if (Utils.randomEvent(prob)) {
			int val = clause.value;
			
			if (val == 0) val++;
			else if (val == clause.getVar().getNumValues()-1) val--;
			else val += Utils.randomBoolean() ? 1 : -1;
			
			clause.value = val;
		}
	}

	
	@Override
	public String toString() {
		StringBuilder str = new StringBuilder();
		for (int i = 0; i < rules.size(); i++) {
			str.append(rules.get(i).toString()).append("\n");
		}
		return str.toString();
	}
	
	
	
	
	public static final Map<Variable, FuzzySet> standardFuzzySets = new HashMap<>();
	public static final Map<Variable, SugenoSet> standardSugenoSets = new HashMap<>();
	
	
	private static void addStandardFuzzySet(FuzzySet set) {
		if (standardFuzzySets.containsKey(set.getVar())) throw new RuntimeException("Set " + set.getName() + " already defined.");
		standardFuzzySets.put(set.getVar(), set);
	}
	
	
	private static void addStandardSugenoSet(SugenoSet set) {
		if (standardSugenoSets.containsKey(set.getVar())) throw new RuntimeException("Set " + set.getName() + " already defined.");
		standardSugenoSets.put(set.getVar(), set);
	}
	
	
	static {
		Variable.load();
		
		FuzzySet hp = new FuzzySet("HP", 0, 1);
		hp.addFuzzyValue("Low", GRADE.createInstance(true, 0.2, 0.4));
		hp.addFuzzyValue("Moderate", TRAPEZOIDAL.createInstance(0.2, 0.4, 0.6, 0.8));
		hp.addFuzzyValue("High", GRADE.createInstance(0.6, 0.8));
		addStandardFuzzySet(hp);
		
		FuzzySet distance = new FuzzySet("Distance", 0, 500);
		distance.addFuzzyValue("Close", GRADE.createInstance(true, 100, 150));
		distance.addFuzzyValue("Medium", TRAPEZOIDAL.createInstance(50, 200, 300, 350));
		distance.addFuzzyValue("Far", TRAPEZOIDAL.createInstance(250, 400, 500, 500));
		distance.addFuzzyValue("OutOfRange", GRADE.createInstance(false, 500, 500));
		distance.normalize();
		addStandardFuzzySet(distance);
		
		FuzzySet direction = new FuzzySet("Direction", -1, 1);
		direction.addFuzzyValue("Left", GRADE.createInstance(true, -0.8, 0));
		direction.addFuzzyValue("Center", TRAPEZOIDAL.createInstance(-0.3, -0.05, 0.05, 0.3));
		direction.addFuzzyValue("Right", GRADE.createInstance(0, 0.8));
		addStandardFuzzySet(direction);
		
		FuzzySet noise = new FuzzySet("Noise", 0, 1);
		noise.addFuzzyValue("None", GRADE.createInstance(true, 0.1, 0.2));
		noise.addFuzzyValue("Quiet", TRAPEZOIDAL.createInstance(0, 0.3, 0.4, 0.7));
		noise.addFuzzyValue("Loud", GRADE.createInstance(0.4, 0.8));
		addStandardFuzzySet(noise);
		
		FuzzySet smell = new FuzzySet("Smell", -1, 1);
		smell.addFuzzyValue("Bad", GRADE.createInstance(true, -0.8, 0.0));
		smell.addFuzzyValue("None", TRAPEZOIDAL.createInstance(-0.2, -0.1, 0.1, 0.2));
		smell.addFuzzyValue("Good", GRADE.createInstance(0.0, 0.8));
		addStandardFuzzySet(smell);
		
		FuzzySet touch = new FuzzySet("Touch", 0, 1);
		touch.addFuzzyValue("None", GRADE.createInstance(true, 0.1, 0.2));
		touch.addFuzzyValue("Light", TRAPEZOIDAL.createInstance(0, 0.4, 0.5, 0.7));
		touch.addFuzzyValue("Hard", GRADE.createInstance(0.5, 0.9));
		addStandardFuzzySet(touch);
		
		FuzzySet foodEaten = new FuzzySet("FoodEaten", -1, 1);
		foodEaten.addFuzzyValue("No", TRAPEZOIDAL.createInstance(-1, -1, 0, 0));
		foodEaten.addFuzzyValue("Yes", TRAPEZOIDAL.createInstance(0, 0, 1, 1));
		addStandardFuzzySet(foodEaten);
		
		SugenoSet move = new SugenoSet("Move");
		move.addValue("Reverse", -1);
		move.addValue("None", 0);
		move.addValue("Forward", 1);
		addStandardSugenoSet(move);
		
		SugenoSet turn = new SugenoSet("Turn");
		turn.addValue("CCV", -1);
		turn.addValue("None", 0);
		turn.addValue("CV", 1);
		addStandardSugenoSet(turn);
		
		SugenoSet shoot = new SugenoSet("Shoot");
		shoot.addValue("No", 0);
		shoot.addValue("Yes", 1);
		addStandardSugenoSet(shoot);
		
		SugenoSet rotate = new SugenoSet("Rotate");
		rotate.addValue("CCV", -1);
		rotate.addValue("None", 0);
		rotate.addValue("CV", 1);
		addStandardSugenoSet(rotate);
		
		SugenoSet push = new SugenoSet("Push");
		push.addValue("No", 0);
		push.addValue("Yes", 1);
		addStandardSugenoSet(push);
	}

	
	@Override
	public BehaviourGenotype treeGenotypeOnePointCrossover(TreeGenotype child, TreeGenotype parent1, TreeGenotype parent2, int splitIndex) {
		throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
	}

	
	@Override
	public BehaviourGenotype treeGenotypeUniformCrossover(TreeGenotype child, TreeGenotype parent1, TreeGenotype parent2) {
		throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
	}

	
}
